# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pathlib

import pandas as pd
import pytest
import sciline as sl
import scipp as sc

from ess.nmx import NMXMcStasWorkflow
from ess.nmx.data import get_small_mcstas
from ess.nmx.mcstas.reduction import merge_panels
from ess.nmx.mcstas.types import (
    DetectorIndex,
    FilePath,
    MaximumCounts,
    NMXRawEventCountsDataGroup,
    NMXReducedDataGroup,
    TimeBinSteps,
)


@pytest.fixture(params=[get_small_mcstas])
def mcstas_file_path(request: pytest.FixtureRequest) -> pathlib.Path:
    return request.param()


@pytest.fixture
def mcstas_workflow(mcstas_file_path: pathlib.Path) -> sl.Pipeline:
    wf = NMXMcStasWorkflow()
    wf[FilePath] = mcstas_file_path
    wf[TimeBinSteps] = 50
    return wf


@pytest.fixture
def multi_bank_mcstas_workflow(mcstas_workflow: sl.Pipeline) -> sl.Pipeline:
    pl = mcstas_workflow.copy()
    pl[NMXReducedDataGroup] = (
        pl[NMXReducedDataGroup]
        .map(pd.DataFrame({DetectorIndex: range(3)}).rename_axis('panel'))
        .reduce(index='panel', func=merge_panels)
    )
    return pl


def test_pipeline_builder(
    mcstas_workflow: sl.Pipeline, mcstas_file_path: pathlib.Path
) -> None:
    assert mcstas_workflow.get(FilePath).compute() == mcstas_file_path


def test_pipeline_mcstas_loader(mcstas_workflow: sl.Pipeline) -> None:
    """Test if the loader graph is complete."""
    mcstas_workflow[DetectorIndex] = 0
    nmx_data = mcstas_workflow.compute(NMXRawEventCountsDataGroup)
    assert isinstance(nmx_data, sc.DataGroup)
    assert nmx_data.sizes['id'] == 1280 * 1280


def test_pipeline_mcstas_reduction(multi_bank_mcstas_workflow: sl.Pipeline) -> None:
    """Test if the loader graph is complete."""
    from scipp.testing import assert_allclose, assert_identical

    from ess.nmx.mcstas import default_parameters

    nmx_reduced_data = multi_bank_mcstas_workflow.compute(NMXReducedDataGroup)
    assert nmx_reduced_data.shape == (3, (1280, 1280)[0] * (1280, 1280)[1], 50)
    # Panel, Pixels, Time bins
    assert isinstance(nmx_reduced_data, sc.DataGroup)

    # Check maximum value of weights.
    assert_allclose(
        nmx_reduced_data['counts'].max().data,
        sc.scalar(default_parameters[MaximumCounts], unit='counts', dtype=float),
        atol=sc.scalar(1e-10, unit='counts'),
        rtol=sc.scalar(1e-8),
    )
    assert_identical(
        nmx_reduced_data['proton_charge'],
        sc.scalar(1e-4, unit='dimensionless')
        * nmx_reduced_data['counts'].data.sum('id').sum('t'),
    )
    assert nmx_reduced_data.sizes['t'] == 50



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# These fixtures cannot be found by pytest,
# if they are not defined in `conftest.py` under `tests` directory.



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import io
import pathlib
import warnings
from typing import Any

import h5py
import numpy as np
import scipp as sc
import scippnexus as snx

from .configurations import Compression
from .types import (
    NMXDetectorMetadata,
    NMXMonitorMetadata,
    NMXProgram,
    NMXSampleMetadata,
    NMXSourceMetadata,
)


def _check_file(
    filename: str | pathlib.Path | io.BytesIO, overwrite: bool
) -> pathlib.Path | io.BytesIO:
    if isinstance(filename, str | pathlib.Path):
        filename = pathlib.Path(filename)
        if filename.exists() and not overwrite:
            raise FileExistsError(
                f"File '{filename}' already exists. Use `overwrite=True` to overwrite."
            )
    return filename


def _create_dataset_from_var(
    *,
    root_entry: h5py.Group,
    var: sc.Variable,
    name: str,
    long_name: str | None = None,
    compression: str | None = None,
    compression_opts: int | tuple[int, int] | None = None,
    chunks: tuple[int, ...] | int | bool | None = None,
    dtype: Any = None,
) -> h5py.Dataset:
    compression_options = {}
    if compression is not None:
        compression_options["compression"] = compression
    if compression_opts is not None:
        compression_options["compression_opts"] = compression_opts

    dataset = root_entry.create_dataset(
        name,
        data=var.values if dtype is None else var.values.astype(dtype, copy=False),
        chunks=chunks,
        **compression_options,
    )
    if var.unit is not None:
        dataset.attrs["units"] = str(var.unit)
    if long_name is not None:
        dataset.attrs["long_name"] = long_name
    return dataset


def _retrieve_compression_arguments(compress_mode: Compression) -> dict:
    """Returns compression filter and opts arguments for the ``compress_mode``.

    Returns an empty dictionary if an unimplemented compression mode
    or `NONE` compression mode is selected.

    """
    if compress_mode == Compression.BITSHUFFLE_LZ4:
        try:
            import bitshuffle.h5

            compression_filter = bitshuffle.h5.H5FILTER
            compression_opts = (0, bitshuffle.h5.H5_COMPRESS_LZ4)
        except ImportError:
            warnings.warn(
                UserWarning(
                    "Could not find the bitshuffle.h5 module from bitshuffle package. "
                    "The bitshuffle package is not installed properly. "
                    "Trying with gzip compression instead..."
                ),
                stacklevel=2,
            )
            compression_filter = "gzip"
            compression_opts = 4
    elif compress_mode == Compression.GZIP:
        compression_filter = "gzip"
        compression_opts = 4
    elif compress_mode == Compression.NONE:
        return {}
    else:
        warnings.warn(
            UserWarning(
                f"Compression Mode {compress_mode} is not implemented yet. "
                "Not Compressing the dataset... "
                "Try `GZIP` or `BITSHUFFLE_LZ4` if compression is needed."
            ),
            stacklevel=2,
        )
        return {}

    return {"compression": compression_filter, "compression_opts": compression_opts}


def _add_arbitrary_metadata(
    nx_entry: h5py.Group, **arbitrary_metadata: sc.Variable
) -> None:
    if not arbitrary_metadata:
        return

    metadata_group = nx_entry.create_group("metadata")
    for key, value in arbitrary_metadata.items():
        if not isinstance(value, sc.Variable):
            import warnings

            msg = f"Skipping metadata key '{key}' as it is not a scipp.Variable."
            warnings.warn(UserWarning(msg), stacklevel=2)
            continue
        else:
            _create_dataset_from_var(
                name=key,
                root_entry=metadata_group,
                var=value,
            )


def _set_default_instrument(nx_entry: snx.Group) -> snx.Group:
    """Return NXinstrument group.

    If 'instrument' exists in the NXentry group, it returns the existing one.
    Otherwise, new NXinstrument group is created and returned.
    The default NXinstrument group has a field 'name' with the instrument name, 'NMX'.
    """
    if "instrument" not in nx_entry:
        nx_instrument = nx_entry.create_class("instrument", 'NXinstrument')
        nx_instrument.create_field(key='name', value='NMX')
    else:
        nx_instrument = nx_entry["instrument"]

    return nx_instrument


def export_static_metadata_as_nxlauetof(
    *,
    sample_metadata: NMXSampleMetadata,
    source_metadata: NMXSourceMetadata,
    program: NMXProgram,
    output_file: str | pathlib.Path | io.BytesIO,
    overwrite: bool = False,
    **arbitrary_metadata: sc.Variable,
) -> None:
    """Export the metadata to a NeXus file with the LAUE_TOF application definition.

    ``Metadata`` in this context refers to the information
    that is not part of the reduced detector counts itself,
    but is necessary for the interpretation of the reduced data.
    Since NMX can have arbitrary number of detectors,
    this function can take multiple detector metadata objects.

    Parameters
    ----------
    sample_metadata:
        Sample metadata object.
    source_metadata:
        Source metadata object.
    monitor_metadata:
        Monitor metadata object.
    output_file:
        Output file path.
    arbitrary_metadata:
        Arbitrary metadata that does not fit into the existing metadata objects.

    """
    _check_file(output_file, overwrite=overwrite)
    with snx.File(output_file, "w") as f:
        f._group.attrs["NX_class"] = "NXlauetof"
        nx_entry = f.create_class(name='entry', class_name='NXlauetof')
        nx_entry.create_field('definitions', value='NXlauetof')
        nx_entry['sample'] = sample_metadata
        nx_entry['reducer'] = program

        nx_instrument = _set_default_instrument(nx_entry)
        nx_instrument['source'] = source_metadata
        _add_arbitrary_metadata(nx_entry._group, **arbitrary_metadata)


def export_monitor_metadata_as_nxlauetof(
    monitor_metadata: NMXMonitorMetadata,
    output_file: str | pathlib.Path | io.BytesIO,
    append_mode: bool = True,
) -> None:
    """Export the detector specific metadata to a NeXus file.

    Since NMX can have arbitrary number of detectors,
    this function can take multiple detector metadata objects.

    Parameters
    ----------
    monitor_metadata:
        Monitor metadata object.
    output_file:
        Output file path.

    """
    if not append_mode:
        raise NotImplementedError("Only append mode is supported for now.")

    with snx.File(output_file, "r+") as f:
        nx_entry = f["entry"]
        nx_entry["control"] = monitor_metadata


def export_detector_metadata_as_nxlauetof(
    detector_metadata: NMXDetectorMetadata,
    output_file: str | pathlib.Path | io.BytesIO,
    append_mode: bool = True,
) -> None:
    """Export the detector specific metadata to a NeXus file.

    Since NMX can have arbitrary number of detectors,
    this function can take multiple detector metadata objects.

    Parameters
    ----------
    detector_metadatas:
        Detector metadata objects.
    output_file:
        Output file path.

    """

    if not append_mode:
        raise NotImplementedError("Only append mode is supported for now.")

    with snx.File(output_file, "r+") as f:
        nx_entry: snx.Group = f["entry"]
        nx_instrument = _set_default_instrument(nx_entry)
        nx_instrument[detector_metadata.detector_name] = detector_metadata


def export_reduced_data_as_nxlauetof(
    detector_name: str,
    da: sc.DataArray,
    output_file: str | pathlib.Path | io.BytesIO,
    *,
    append_mode: bool = True,
    compress_mode: Compression = Compression.BITSHUFFLE_LZ4,
) -> None:
    """Export the reduced data to a NeXus file with the LAUE_TOF application definition.

    Even though this function only exports
    reduced data(detector counts and its coordinates),
    the input should contain all the necessary metadata
    for minimum sanity check.

    Parameters
    ----------
    dg:
        Reduced data and metadata.
    output_file:
        Output file path.
    append_mode:
        If ``True``, the file is opened in append mode.
        If ``False``, the file is opened in None-append mode.
        > None-append mode is not supported for now.
        > Only append mode is supported for now.
    compress_mode:
        The detector counts are compressed using the ``compress_mode``.
        It is because only the detector counts are expected to be large.
        If ``Compression.BITSHUFFLE_LZ4`` is selected
        but the bitshuffle is not supported for the environment,
        it will fall back to ``Compression.GZIP``.
        Select ``Compression.NONE`` if compression is not needed.

    """
    if not append_mode:
        raise NotImplementedError("Only append mode is supported for now.")

    with h5py.File(output_file, "r+") as f:
        nx_detector: h5py.Group = f[f"entry/instrument/{detector_name}"]
        # Data - shape: [n_x_pixels, n_y_pixels, n_tof_bins]
        # The actual application definition defines it as integer,
        # so we overwrite the dtype here.

        compression_args = _retrieve_compression_arguments(compress_mode)
        if compress_mode != Compression.NONE:  # Calculate the chunk sizes
            num_x, num_y = da.sizes['x_pixel_offset'], da.sizes['y_pixel_offset']
            compression_args['chunks'] = (num_x, num_y, 1)  # Chunk along tof axis

        data_dset = _create_dataset_from_var(
            name="data",
            root_entry=nx_detector,
            var=da.data,
            dtype=np.uint,
            **compression_args,
        )

        data_dset.attrs["signal"] = 1
        data_dset.attrs["axes"] = list(da.dims)

        if 'tof' in da.coords:
            time_field_name = "time_of_flight"
            time_coord_name = "tof"
            time_dim = "tof"
        elif 'event_time_offset' in da.coords:
            time_field_name = "event_time_offset"
            time_coord_name = "event_time_offset"
            time_dim = "event_time_offset"
        else:
            raise ValueError("Could not find time-related bin edges to store.")

        _create_dataset_from_var(
            name=time_field_name,
            root_entry=nx_detector,
            var=sc.midpoints(da.coords[time_coord_name], dim=time_dim),
        )

        _create_dataset_from_var(
            name='original_time_edges',
            root_entry=nx_detector,
            var=da.coords[time_coord_name],
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import argparse
import enum
import glob
import logging
import pathlib
import sys
from functools import partial
from types import UnionType
from typing import Literal, TypeGuard, TypeVar, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from .configurations import InputConfig, OutputConfig, ReductionConfig, WorkflowConfig


def _validate_annotation(annotation) -> TypeGuard[type]:
    def _validate_atomic_type(annotation) -> bool:
        return (
            (annotation in (int, float, str, bool))
            or (isinstance(annotation, type) and issubclass(annotation, enum.StrEnum))
            or (get_origin(annotation) is Literal)
        )

    return (
        _validate_atomic_type(annotation)
        or (
            (origin := get_origin(annotation)) in (Union, UnionType)
            and _validate_atomic_type(_get_no_nonetype_args(annotation))
        )
        or (
            origin in (list, tuple, set)
            and len(args := get_args(annotation)) > 0
            and _validate_atomic_type(args[0])
        )
    )


def _get_no_nonetype_args(annotation) -> type:
    origin_type = get_origin(annotation)
    if (origin_type is UnionType or origin_type is Union) and type(None) in (
        union_args := get_args(annotation)
    ):
        arg_types = set(union_args) - {type(None)}
        if len(arg_types) > 1:
            raise TypeError(
                "Optional type with single non-None type is not supported: "
                f"{annotation}"
            )
        return next(iter(arg_types))
    return annotation


def _is_appendable_type(annotation) -> bool:
    return get_origin(annotation) in (list, tuple, set)


def _retrieve_field_value(
    field_name: str, field_info: FieldInfo, args: argparse.Namespace
):
    if isinstance(field_info.annotation, type) and issubclass(
        field_info.annotation, enum.StrEnum
    ):
        return field_info.annotation[getattr(args, field_name)]
    return getattr(args, field_name)


def add_args_from_pydantic_model(
    *, model_cls: type[BaseModel], parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """Add arguments to the parser from the pydantic model class.

    Each field in the model class is added as a command line argument
    with the name `--{field-name}`.
    Arguments are added based on fields' information:
      - type annotation (type, choices, nargs)
      - description (help text)
      - default value (default, required and help text)

    Supported annotation for command arguments:
      - Atomic types: int, float, str, bool, enum.StrEnum, Literal
      - Optional[AtomicType]
      - List[AtomicType], Tuple[AtomicType, ...], Set[AtomicType]

    Parameters
    ----------
    model_cls:
        Pydantic model class to extract the arguments from.
    parser:
        Argument parser to add the arguments to.
        It adds a new argument group for the model.
        The group name is taken from the model's title config if available,
        otherwise the model class name is used.

    """
    group = parser.add_argument_group(
        model_cls.model_config.get("title", model_cls.__name__)
    )
    for field_name, field_info in model_cls.model_fields.items():
        add_argument = partial(group.add_argument, f"--{field_name.replace('_', '-')}")

        if not _validate_annotation(field_info.annotation):
            raise TypeError(f"Unsupported annotation type: {field_info.annotation}")

        arg_type = _get_no_nonetype_args(field_info.annotation)
        if _is_appendable_type(arg_type):
            nargs = '+'
            arg_type = get_args(field_info.annotation)[0]
        else:
            nargs = None
            arg_type = arg_type

        required = field_info.default is PydanticUndefined
        default = ... if required else field_info.default

        if arg_type is bool:
            add_argument = partial(add_argument, action='store_true')
        elif isinstance(arg_type, type) and issubclass(arg_type, enum.StrEnum):
            add_argument = partial(
                add_argument,
                type=str,
                choices=[str(e) for e in arg_type],
            )
            default = default.name if isinstance(default, enum.StrEnum) else default
        elif get_origin(arg_type) is Literal:
            add_argument = partial(
                add_argument,
                type=str,
                choices=[str(lit) for lit in get_args(arg_type)],
            )
        else:
            add_argument = partial(add_argument, type=arg_type, nargs=nargs)

        help_text = ' '.join([field_info.description or '', f"(default: {default})"])
        add_argument(default=default, required=required, help=help_text)

    return parser


T = TypeVar('T', bound=BaseModel)


def from_args(cls: type[T], args: argparse.Namespace) -> T:
    """Create an instance of the pydantic model from the argparse namespace.

    It ignores any extra arguments in the namespace that are not part of the model.
    """
    kwargs = {
        field_name: _retrieve_field_value(field_name, field_info, args)
        for field_name, field_info in cls.model_fields.items()
    }
    return cls(**kwargs)


def build_reduction_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Command line arguments for the ESS NMX reduction. "
        "It assumes 14 Hz pulse speed."
    )
    parser = add_args_from_pydantic_model(model_cls=InputConfig, parser=parser)
    parser = add_args_from_pydantic_model(model_cls=WorkflowConfig, parser=parser)
    parser = add_args_from_pydantic_model(model_cls=OutputConfig, parser=parser)
    return parser


def reduction_config_from_args(args: argparse.Namespace) -> ReductionConfig:
    return ReductionConfig(
        inputs=from_args(InputConfig, args),
        workflow=from_args(WorkflowConfig, args),
        output=from_args(OutputConfig, args),
    )


def build_logger(args: argparse.Namespace | OutputConfig) -> logging.Logger:
    logger = logging.getLogger(__name__)
    if args.verbose:
        logger.setLevel(logging.INFO)
        logger.addHandler(logging.StreamHandler(sys.stdout))
    return logger


def collect_matching_input_files(*input_file_patterns: str) -> list[pathlib.Path]:
    """Helper to collect input files matching the given patterns."""

    input_files: list[str] = []
    for pattern in input_file_patterns:
        input_files.extend(glob.glob(pattern))

    # Remove duplicates and sort
    return sorted({pathlib.Path(f).resolve() for f in input_files})



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import NewType, TypeVar

import scipp as sc

from .mtz_io import NMXMtzDataArray

# User defined or configurable types
WavelengthBins = NewType("WavelengthBins", sc.Variable | int)
"""User configurable wavelength binning"""
ReferenceWavelength = NewType("ReferenceWavelength", sc.Variable | None)
"""The wavelength to select reference intensities."""

# Computed types
"""Filtered mtz dataframe by the quad root of the sample standard deviation."""
WavelengthBinned = NewType("WavelengthBinned", sc.DataArray)
"""Binned mtz dataframe by wavelength(LAMBDA) with derived columns."""
SelectedReferenceWavelength = NewType("SelectedReferenceWavelength", sc.Variable)
"""The wavelength to select reference intensities."""
ReferenceIntensities = NewType("ReferenceIntensities", sc.DataArray)
"""Reference intensities selected by the wavelength."""
EstimatedScaleFactor = NewType("EstimatedScaleFactor", sc.DataArray)
"""The estimated scale factor from the reference intensities per ``hkl_asu``."""
EstimatedScaledIntensities = NewType("EstimatedScaledIntensities", sc.DataArray)
"""Scaled intensities by the estimated scale factor."""
FilteredEstimatedScaledIntensities = NewType(
    "FilteredEstimatedScaledIntensities", sc.DataArray
)

T = TypeVar("T")


def get_wavelength_binned(
    mtz_da: NMXMtzDataArray,
    wavelength_bins: WavelengthBins,
) -> WavelengthBinned:
    """Bin the whole dataset by wavelength(LAMBDA).

    Parameters
    ----------
    mtz_da:
        The merged dataset.

    wavelength_bins:
        The wavelength(LAMBDA) bins.

    Notes
    -----
        Wavelength(LAMBDA) binning should always be done on the merged dataset.

    """
    return WavelengthBinned(mtz_da.bin({"wavelength": wavelength_bins}))


def _is_bin_empty(binned: sc.DataArray, idx: int) -> bool:
    """Check if the bin is empty."""
    return binned[idx].values.size == 0


def _get_middle_bin_idx(binned: sc.DataArray) -> int:
    """Find the middle bin index.

    If the middle one is empty, the function will search for the nearest.
    """
    middle_number, offset = len(binned) // 2, 0

    while 0 < (cur_idx := middle_number + offset) < len(binned) and _is_bin_empty(
        binned, cur_idx
    ):
        offset = -offset + 1 if offset <= 0 else -offset

    if _is_bin_empty(binned, cur_idx):
        raise ValueError("No reference group found.")

    return cur_idx


def get_reference_wavelength(
    binned: WavelengthBinned,
    reference_wavelength: ReferenceWavelength,
) -> SelectedReferenceWavelength:
    """Select the reference wavelength.

    Parameters
    ----------
    binned:
        The wavelength binned data.

    reference_wavelength:
        The reference wavelength to select the intensities.
        If ``None``, the middle group is selected.
        It should be a scalar variable as it is selecting one of bins.

    """
    if reference_wavelength is None:
        ref_idx = _get_middle_bin_idx(binned)
        return SelectedReferenceWavelength(binned.coords["wavelength"][ref_idx])
    else:
        return SelectedReferenceWavelength(reference_wavelength)


def get_reference_intensities(
    binned: WavelengthBinned,
    reference_wavelength: SelectedReferenceWavelength,
) -> ReferenceIntensities:
    """Find the reference intensities by the wavelength.

    Parameters
    ----------
    binned:
        The wavelength binned data.

    reference_wavelength:
        The reference wavelength to select the intensities.

    Raises
    ------
    ValueError:
        If no reference group is found.

    """
    if reference_wavelength is None:
        ref_idx = _get_middle_bin_idx(binned)
        return binned[ref_idx].values.copy(deep=False)
    else:
        if reference_wavelength.dims:
            raise ValueError("Reference wavelength should be a scalar.")
        try:
            return binned["wavelength", reference_wavelength].values.copy(deep=False)
        except IndexError as err:
            raise IndexError(f"{reference_wavelength} out of range.") from err


def estimate_scale_factor_per_hkl_asu_from_reference(
    reference_intensities: ReferenceIntensities,
) -> EstimatedScaleFactor:
    """Calculate the estimated scale factor per ``hkl_asu``.

    The estimated scale factor is calculated as the average
    of the inverse of the non-empty reference intensities.

    It is part of the calculation of estimated scaled intensities
    for fitting the scaling model.

    .. math::

        EstimatedScaleFactor_{(hkl)} = \\dfrac{
            \\sum_{i=1}^{N_{(hkl)}} \\dfrac{1}{I_{i}}
        }{
            N_{(hkl)}
        }
        = average( \\dfrac{1}{I_{(hkl)}} )

    Estimated scale factor is calculated per ``hkl_asu``.
    This is part of the calculation of roughly-scaled-intensities
    for fitting the scaling model.
    The whole procedure is described in :func:`average_roughly_scaled_intensities`.

    Parameters
    ----------
    reference_intensities:
        The reference intensities selected by wavelength.

    Returns
    -------
    :
        The estimated scale factor per ``hkl_asu``.
        The result should have a dimension of ``hkl_asu``.

        It does not have a dimension of ``wavelength`` since
        it is calculated from the reference intensities,
        which is selected by one ``wavelength``.

    """
    # Workaround for https://github.com/scipp/scipp/issues/3046
    # and https://github.com/scipp/scipp/issues/3425
    import numpy as np

    unique_hkl = np.unique(reference_intensities.coords["hkl_asu"].values)
    group_var = sc.array(dims=["hkl_asu"], values=unique_hkl)
    grouped = reference_intensities.group(group_var)

    return EstimatedScaleFactor((1 / grouped).bins.mean())


def average_roughly_scaled_intensities(
    binned: WavelengthBinned,
    scale_factor: EstimatedScaleFactor,
) -> EstimatedScaledIntensities:
    """Scale the intensities by the estimated scale factor.

    Parameters
    ----------
    binned:
        Intensities binned in the wavelength dimension.
        It will be grouped by reflection (hkl) in the process.

    scale_factor:
        The estimated scale factor per reflection(hkl) of the reference wavelength bin.
        See :func:`estimate_scale_factor_per_hkl_asu_from_reference`
        for the calculation of the estimated scale factor.

        .. math::

            EstimatedScaleFactor_{(hkl)} =
            average( \\dfrac{1}{I_{\\lambda=reference, (hkl)}} )

    Returns
    -------
    :
        Average scaled intensities on ``hkl(asu)`` indices per wavelength.

    Notes
    -----
    The average of roughly scaled intensities are calculated by the following formula:

    .. math::

        EstimatedScaledI_{\\lambda}
        = \\dfrac{
            \\sum_{i=1}^{N_{\\lambda, (hkl)}}
            EstimatedScaledI_{\\lambda, (hkl)}
        }{
            N_{\\lambda, (hkl)}
        }

    And scaled intensities on each ``hkl(asu)`` indices per wavelength
    are calculated by the following formula:

    .. math::
        :nowrap:

        \\begin{eqnarray}
        EstimatedScaledI_{\\lambda, (hkl)} \\\\
        = \\dfrac{
            \\sum_{i=1}^{N_{\\lambda=reference, (hkl)}}
            \\sum_{j=1}^{N_{\\lambda, (hkl)}}
            \\dfrac{I_{j}}{I_{i}}
        }{
            N_{\\lambda=reference, (hkl)}*N_{\\lambda, (hkl)}
        } \\\\
        = \\dfrac{
            \\sum_{i=1}^{N_{\\lambda=reference, (hkl)}} \\dfrac{1}{I_{i}}
        }{
            N_{\\lambda=reference, (hkl)}
        } * \\dfrac{
            \\sum_{j=1}^{N_{\\lambda, (hkl)}} I_{j}
        }{
            N_{\\lambda, (hkl)}
        } \\\\
        = average( \\dfrac{1}{I_{\\lambda=reference, (hkl)}} )
        * average( I_{\\lambda, (hkl)} )
        \\end{eqnarray}

    Therefore the ``binned(wavelength dimension)`` should be
    grouped along the ``hkl(asu)`` coordinate in the calculation.

    """
    # Group by HKL_EQ of the estimated scale factor from reference intensities
    grouped = binned.group(scale_factor.coords["hkl_asu"])

    # Drop variances of the scale factor
    # Scale each group each bin by the scale factor
    intensities = sc.nanmean(
        grouped.bins.nanmean() * sc.values(scale_factor), dim="hkl_asu"
    )
    # Take the midpoints of the wavelength bin coordinates
    # to represent the average wavelength of the bin
    # It is because the bin-edges are dropped while flattening the data
    # and the data is expected to be filtered after this step.
    intensities.coords["wavelength"] = sc.midpoints(
        intensities.coords["wavelength"],
    )
    return EstimatedScaledIntensities(intensities)


ScaledIntensityLeftTailThreshold = NewType(
    "ScaledIntensityLeftTailThreshold", sc.Variable
)
"""The threshold to cut the left tail of the estimated scaled intensities."""
DEFAULT_LEFT_TAIL_THRESHOLD = ScaledIntensityLeftTailThreshold(sc.scalar(0.1))
ScaledIntensityRightTailThreshold = NewType(
    "ScaledIntensityRightTailThreshold", sc.Variable
)
"""The threshold to cut the right tail of the estimated scaled intensities."""
DEFAULT_RIGHT_TAIL_THRESHOLD = ScaledIntensityRightTailThreshold(sc.scalar(2.0))


def cut_tails(
    scaled_intensities: EstimatedScaledIntensities,
    left_threashold: ScaledIntensityLeftTailThreshold = DEFAULT_LEFT_TAIL_THRESHOLD,
    right_threshold: ScaledIntensityRightTailThreshold = DEFAULT_RIGHT_TAIL_THRESHOLD,
) -> FilteredEstimatedScaledIntensities:
    """Cut the right tail of the estimated scaled intensities by the threshold.

    Parameters
    ----------
    scaled_intensities:
        The scaled intensities to be filtered.

    left_threashold:
        The threshold to be cut from the left tail.

    right_threshold:
        The threshold to be cut from the right tail.

    Returns
    -------
    :
        The filtered scaled intensities with the tails cut.

    """
    return FilteredEstimatedScaledIntensities(
        scaled_intensities[
            (scaled_intensities.data > left_threashold)
            & (scaled_intensities.data < right_threshold)
        ].copy(deep=False)
    )


@dataclass
class FittingResult:
    """Result of the fitting process."""

    fitting_func: Callable[..., sc.DataArray]
    """The fitting function to be used for fitting."""
    params: Mapping
    """Parameters of the fitting function."""
    covariance: Mapping
    """Covariance of the :attr:`~FittingParams`."""
    fit_output: sc.DataArray
    """The final output of the fitting function."""


def polyval_wavelength(
    wavelength: sc.Variable, *, out_unit: str, **kwargs
) -> sc.DataArray:
    """Polynomial helper for fitting.

    The coefficients are adjusted to make the fitting result
    have ``out_unit`` as unit.

    Parameters
    ----------
    wavelength:
        The wavelength coordinate.
    out_unit:
        The unit of the output.
    **kwargs:
        The polynomial coefficients.

    Returns
    -------
    :
        The polynomial calculated at the wavelength.


    """
    # We need to use float64 precision because
    # the curve fit routine depends on finite difference
    # estimates of the derivative.
    wavelength = wavelength.to(dtype='float64')
    out = sc.zeros_like(wavelength)
    out.unit = out_unit
    xk = sc.ones_like(wavelength)
    for _, arg_value in enumerate(kwargs.values()):
        out += sc.values(arg_value) * xk * sc.scalar(1.0, unit=out.unit / xk.unit)
        xk *= wavelength
    return out


WavelengthFittingPolynomialDegree = NewType("WavelengthFittingPolynomialDegree", int)
DEFAULT_WAVELENGTH_FITTING_POLYNOMIAL_DEGREE = WavelengthFittingPolynomialDegree(7)


def fit_wavelength_scale_factor_polynomial(
    estimated_intensities: FilteredEstimatedScaledIntensities,
    *,
    n_degree: WavelengthFittingPolynomialDegree,
) -> FittingResult:
    """Fit the wavelength scale factor polynomial.

    It uses :func:`polyval_wavelength` as the fitting function
    and :func:`scipp.optimize.curve_fit` for the fitting process.
    The initial guess for the polynomial coefficients is set to 1
    for all degrees.
    The unit of the coefficients is adjusted to make the fitting result
    dimensionless.

    Parameters
    ----------
    estimated_intensities:
        The estimated scaled intensities to be fitted.
    n_degree:
        The degree of the polynomial to be fitted.

    Returns
    -------
    :
        The fitting result.

    """

    from functools import partial

    fitting_func = partial(polyval_wavelength, out_unit="dimensionless")
    p_result, cov_result = sc.curve_fit(
        coords=["wavelength"],
        f=fitting_func,
        da=estimated_intensities,
        p0={f"arg{i}": sc.scalar(1) for i in range(n_degree)},
    )
    data = fitting_func(estimated_intensities.coords["wavelength"], **p_result)
    return FittingResult(
        fitting_func=fitting_func,
        params=p_result,
        covariance=cov_result,
        fit_output=sc.DataArray(
            data=data.data,
            coords={"wavelength": estimated_intensities.coords["wavelength"]},
        ),
    )


WavelengthScaleFactors = NewType("WavelengthScaleFactors", sc.DataArray)
"""The scale factors of `"wavelength"`."""


def calculate_wavelength_scale_factor(
    fitting_result: FittingResult,
    reference_wavelength: SelectedReferenceWavelength,
) -> WavelengthScaleFactors:
    """Calculate the scale factors along the `"wavelength"`."""

    scaled_reference = fitting_result.fitting_func(
        reference_wavelength, **fitting_result.params
    )
    scale_factor = fitting_result.fit_output / scaled_reference
    return WavelengthScaleFactors(scale_factor)


providers = (
    cut_tails,
    get_wavelength_binned,
    get_reference_wavelength,
    get_reference_intensities,
    estimate_scale_factor_per_hkl_asu_from_reference,
    average_roughly_scaled_intensities,
    fit_wavelength_scale_factor_polynomial,
    calculate_wavelength_scale_factor,
)
"""Providers for scaling data."""

default_parameters = {
    WavelengthBins: sc.linspace("wavelength", 1.8, 3.5, 250, unit="angstrom"),
    ScaledIntensityLeftTailThreshold: DEFAULT_LEFT_TAIL_THRESHOLD,
    ScaledIntensityRightTailThreshold: DEFAULT_RIGHT_TAIL_THRESHOLD,
    WavelengthFittingPolynomialDegree: WavelengthFittingPolynomialDegree(7),
}
"""Default parameters for scaling data."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Scipp contributors (https://github.com/scipp)
import warnings

import scipp as sc
import scippnexus as snx
from ess.reduce.nexus.types import FilePath, NeXusFile
from scippneutron.metadata import RadiationProbe, SourceType

from .types import ControlMode


def _validate_entry(entry: snx.Group) -> None:
    if str(entry.attrs['NX_class']) != 'NXlauetof':
        raise ValueError("File entry is not NXlauetof.")
    _MANDATORY_FIELDS = ('control', 'instrument', 'sample')
    missing_fields = [field for field in _MANDATORY_FIELDS if field not in entry]
    if any(missing_fields):
        raise ValueError("File entry missing mandatory fields, ", missing_fields)


def _as_vector(var: sc.Variable) -> sc.Variable:
    if var.dims == () and var.dtype == sc.DType.vector3:
        return var
    elif len(var.dims) == 1 and var.sizes[var.dim] == 3:
        return sc.vector(value=var.values, unit=var.unit)
    else:
        warnings.warn(
            f"Cannot convert to vector3 scalar: {var}. "
            "Falling back to the original form.",
            UserWarning,
            stacklevel=3,
        )
        return var


def _handle_sample(sample_dg: sc.DataGroup, sample: snx.Group) -> sc.DataGroup:
    sample_dg['crystal_rotation'] = _as_vector(sample_dg['crystal_rotation'])
    sample_dg['position'] = _as_vector(sample_dg['position'])
    unit_cell = sample_dg.pop('unit_cell')
    sample_dg['unit_cell_length'] = sc.vector(
        unit_cell[:3], unit=sample['unit_cell'].attrs['length-unit']
    )
    sample_dg['unit_cell_angle'] = sc.vector(
        unit_cell[3:], unit=sample['unit_cell'].attrs['angle-unit']
    )
    return sample_dg


def _handle_monitor(control_dg: sc.DataGroup, control: snx.Group) -> sc.DataGroup:
    tof_bin_coord_key = 'tof_bin_coord'

    if tof_bin_coord_key in control.attrs:
        tof_bin_coord = control.attrs['tof_bin_coord']
        control_dg['tof_bin_coord'] = tof_bin_coord
        data: sc.DataArray = control_dg['data']
        data.coords[tof_bin_coord] = data.coords.pop('time_of_flight')

    control_dg['mode'] = ControlMode[control_dg['mode']]

    return control_dg


def _handle_source(instrument_dg: sc.DataGroup, instrument: snx.Group) -> sc.DataGroup:
    source_dg = instrument_dg['source']
    distance = source_dg.pop('distance')
    position = sc.vector(
        instrument['source']['distance'].attrs['position'], unit=distance.unit
    )
    source_dg['position'] = position
    source_dg['source_type'] = SourceType(source_dg.pop('type'))
    source_dg['probe'] = RadiationProbe(source_dg['probe'])


def _restore_positions(
    *, metadatas: sc.DataGroup, fast_axis_dim: str, slow_axis_dim: str, sizes: dict
) -> sc.Variable:
    fast_axis = metadatas['fast_axis']
    fast_axis_size = sizes[fast_axis_dim]
    slow_axis = metadatas['slow_axis']
    slow_axis_size = sizes[slow_axis_dim]

    pixel_sizes = {
        'x_pixel_offset': metadatas['x_pixel_size'],
        'y_pixel_offset': metadatas['y_pixel_size'],
    }

    fast_axis_offsets = (
        sc.arange(dim=fast_axis_dim, start=0.0, stop=fast_axis_size)
        * pixel_sizes[fast_axis_dim]
        * fast_axis
    )
    slow_axis_offsets = (
        sc.arange(dim=slow_axis_dim, start=0.0, stop=slow_axis_size)
        * pixel_sizes[slow_axis_dim]
        * slow_axis
    )
    # The slow axis should be the outer most dimension.
    detector_sizes = {slow_axis_dim: slow_axis_size, fast_axis_dim: fast_axis_size}

    pixel_offsets = fast_axis_offsets.broadcast(
        sizes=detector_sizes
    ) + slow_axis_offsets.broadcast(sizes=detector_sizes)

    detetor_center = metadatas['origin']
    slow_axis_width = pixel_sizes[slow_axis_dim] * slow_axis_size
    fast_axis_width = pixel_sizes[fast_axis_dim] * fast_axis_size
    detector_corner = (
        detetor_center
        - (slow_axis_width / 2) * slow_axis
        + pixel_sizes[slow_axis_dim] * slow_axis / 2
        - (fast_axis_width / 2) * fast_axis
        + pixel_sizes[fast_axis_dim] * fast_axis / 2
    )

    return pixel_offsets + detector_corner


def _handle_detector_data(
    instrument_dg: sc.DataGroup, instrument: snx.Group
) -> sc.DataGroup:
    detectors: sc.DataGroup[sc.DataGroup] = sc.DataGroup(
        {
            det_name: instrument_dg.pop(det_name)
            for det_name in instrument[snx.NXdetector].keys()
        }
    )
    instrument_dg['detectors'] = detectors
    time_coord_name = next(iter({'tof', 'event_time_offset'} & set(detectors.dims)))
    time_field_name = 'time_of_flight' if time_coord_name == 'tof' else time_coord_name

    for det_name, det_gr in detectors.items():
        # These fields are part of the histogram as data and coordinate.
        non_meta_keys = ('data', 'time_of_flight', 'event_time_offset')
        all_keys = list(filter(lambda key: key not in non_meta_keys, det_gr.keys()))
        metadatas = sc.DataGroup({key: det_gr.pop(key) for key in all_keys})

        for vector_field in ('slow_axis', 'fast_axis', 'origin'):
            metadatas[vector_field] = _as_vector(metadatas[vector_field])

        det_gr['metadata'] = metadatas
        fast_axis_dim = instrument[det_name]['fast_axis'].attrs['dim']
        slow_axis_dim = instrument[det_name]['slow_axis'].attrs['dim']
        metadatas['fast_axis_dim'] = fast_axis_dim
        metadatas['slow_axis_dim'] = slow_axis_dim
        metadatas['detector_name'] = det_name
        metadatas['first_pixel_position'] = sc.vector(
            instrument[det_name]['origin'].attrs['first_pixel_position'],
            unit=metadatas['origin'].unit,
        )
        time_coord = metadatas.pop('original_time_edges')
        mid_time = det_gr.pop(time_field_name)
        if sc.any(sc.midpoints(time_coord) != mid_time):
            warnings.warn(
                "Time bin edges and mid point coordinates do not agree.",
                UserWarning,
                stacklevel=3,
            )
        det_gr['data'] = sc.DataArray(
            data=det_gr['data'],
            coords={
                time_coord_name: time_coord,
                'position': _restore_positions(
                    metadatas=metadatas,
                    fast_axis_dim=fast_axis_dim,
                    slow_axis_dim=slow_axis_dim,
                    sizes=det_gr['data'].sizes,
                ),
            },
        )


def load_essnmx_nxlauetof(file: str | FilePath | NeXusFile) -> sc.DataGroup:
    with snx.File(file, mode='r') as f:
        with warnings.catch_warnings(action='ignore'):
            # Expecting warnings for loading NXdetectors.
            # The data array reconstruction is handled manually later.
            dg = f[()]

        _validate_entry(entry := f['entry'])
        _handle_sample(dg['entry']['sample'], entry['sample'])
        _handle_monitor(dg['entry']['control'], entry['control'])
        _handle_source(dg['entry']['instrument'], entry['instrument'])
        _handle_detector_data(dg['entry']['instrument'], entry['instrument'])

    return dg['entry']



from dataclasses import dataclass
from typing import Any, NewType

import scipp as sc

FilePath = NewType("FilePath", str)
"""File name of a file containing the results of a McStas run"""

DetectorIndex = NewType("DetectorIndex", int | sc.Variable | sc.DataArray)
"""Index of the detector to load. Index ordered by the id:s of the pixels"""

DetectorName = NewType("DetectorName", str)
"""Name of the detector to load"""

DetectorBankPrefix = NewType("DetectorBankPrefix", str)
"""Prefix identifying the event data array containing
the events from the selected detector"""

MaximumCounts = NewType("MaximumCounts", int)
"""Maximum number of counts after scaling the event counts"""

MaximumProbability = NewType("MaximumProbability", sc.Variable)
"""Maximum probability to scale the McStas event counts"""

McStasWeight2CountScaleFactor = NewType("McStasWeight2CountScaleFactor", sc.Variable)
"""Scale factor to convert McStas weights to counts"""

NMXExperimentMetadata = NewType("NMXExperimentMetadata", sc.DataGroup)
"""Metadata of the experiment"""

NMXDetectorMetadata = NewType("NMXDetectorMetadata", sc.DataGroup)
"""Metadata of the detector"""

RawEventProbability = NewType("RawEventProbability", sc.DataArray)
"""DataArray containing the event probabilities read from the McStas file,
has coordinates 'id' and 't' """

NMXRawEventCountsDataGroup = NewType("NMXRawEventCountsDataGroup", sc.DataGroup)
"""DataGroup containing the RawEventData, experiment metadata and detector metadata"""

ProtonCharge = NewType("ProtonCharge", sc.Variable)
"""The proton charge signal"""

CrystalRotation = NewType("CrystalRotation", sc.Variable)
"""Rotation of the crystal"""

DetectorGeometry = NewType("DetectorGeometry", Any)
"""Description of the geometry of the detector banks"""

TimeBinSteps = NewType("TimeBinSteps", int)
"""Number of bins in the binning of the time coordinate"""

PixelIds = NewType("PixelIds", sc.Variable)
"""The pixel ids of the detector"""

NMXReducedProbability = NewType("NMXReducedProbability", sc.DataArray)
"""Histogram of time-of-arrival and pixel-id."""

NMXReducedCounts = NewType("NMXReducedCounts", sc.DataArray)
"""Histogram of time-of-arrival and pixel-id."""

NMXReducedDataGroup = NewType("NMXReducedDataGroup", sc.DataGroup)
"""Datagroup containing Histogram(id, t), experiment metadata and detector metadata"""

MinimumTimeOfArrival = NewType("MinimumTimeOfArrival", sc.Variable)
"""Minimum time of arrival of the raw data"""

MaximumTimeOfArrival = NewType("MaximumTimeOfArrival", sc.Variable)
"""Maximum time of arrival of the raw data"""


@dataclass
class NMXRawDataMetadata:
    """Metadata of the raw data, i.e. maximum weight and min/max time of arrival"""

    max_probability: MaximumProbability
    min_toa: MinimumTimeOfArrival
    max_toa: MaximumTimeOfArrival



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pathlib

from ess.reduce.data import Entry, make_registry

_version = "1"

__all__ = [
    "get_path",
    "get_small_mcstas",
    "get_small_mtz_samples",
    "get_small_nmx_nexus",
    "get_small_random_mtz_samples",
]


_registry = make_registry(
    "ess/nmx",
    version=_version,
    files={
        "small_mcstas_sample.h5": "md5:2afaac205d13ee857ee5364e3f1957a7",
        "mtz_samples.tar.gz": Entry(
            alg="md5", chk="bed1eaf604bbe8725c1f6a20ca79fcc0", extractor="untar"
        ),
        "mtz_random_samples.tar.gz": Entry(
            alg="md5", chk="c8259ae2e605560ab88959e7109613b6", extractor="untar"
        ),
        "small_nmx_nexus.hdf.zip": Entry(
            alg="md5", chk="96877cddc9f6392c96890069657710ca", extractor="unzip"
        ),
    },
)


def get_small_mcstas() -> pathlib.Path:
    """McStas file that contains only ``bank0(1-3)`` in the ``data`` group.

    Real McStas file should contain more dataset under ``data`` group.
    McStas version >=3.
    """
    return get_path("small_mcstas_sample.h5")


def get_path(name: str) -> pathlib.Path:
    """
    Return the path to a data file bundled with ess nmx.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _registry.get_path(name)


def get_small_mtz_samples() -> list[pathlib.Path]:
    """Return a list of path to MTZ sample files randomly chosen from real dataset.

    This samples also contain optional columns.
    """
    return _registry.get_paths("mtz_samples.tar.gz")


def get_small_random_mtz_samples() -> list[pathlib.Path]:
    """Return a list of path to MTZ sample files filled with random values

    This sample only contains mandatory columns for the workflow examples.
    They are made for documentation, not necessarily for testing.
    Use ``get_small_mtz_samples`` for testing since they are
    more representative of real data.
    """
    return _registry.get_paths("mtz_random_samples.tar.gz")


def get_small_nmx_nexus() -> pathlib.Path:
    """Return the path to a small NMX NeXus file."""

    return get_path("small_nmx_nexus.hdf.zip")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import io

import numpy as np
import pytest
import scipp as sc

from ess.nmx.mcstas.nexus import export_as_nexus
from ess.nmx.mcstas.types import NMXReducedDataGroup


@pytest.fixture
def reduced_data() -> NMXReducedDataGroup:
    rng = np.random.default_rng(42)
    id_list = sc.array(dims=['event'], values=rng.integers(0, 12, size=100))
    t_list = sc.array(dims=['event'], values=rng.random(size=100, dtype=float))
    counts = (
        sc.DataArray(
            data=sc.ones(dims=['event'], shape=[100]),
            coords={'id': id_list, 't': t_list},
        )
        .group('id')
        .hist(t=10)
    )

    return NMXReducedDataGroup(
        sc.DataGroup(
            dict(  # noqa: C408
                counts=counts,
                proton_charge=sc.scalar(1.0, unit='counts'),
                crystal_rotation=sc.vector(value=[0.0, 20.0, 0.0], unit='deg'),
                fast_axis=sc.vectors(
                    dims=['panel'],
                    values=[[1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
                    unit='m',
                ),
                slow_axis=sc.vectors(
                    dims=['panel'],
                    values=[[0.0, 1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 0.0]],
                    unit='m',
                ),
                origin_position=sc.vectors(
                    dims=['panel'],
                    values=[[-0.2, 0.0, 0.0], [0.0, 0.0, 0.0], [0.2, 0.0, 0.0]],
                    unit='m',
                ),
                sample_position=sc.vector(value=[0.0, 0.0, 0.0], unit='m'),
                source_position=sc.vector(value=[-3, 0.0, -4], unit='m'),
                sample_name=sc.scalar('Unit Test Sample'),
                position=sc.zeros(dims=['panel', 'id'], shape=[3, 4], unit='m'),
            )
        )
    )


def _is_bitshuffle_available() -> bool:
    import platform

    return not (
        platform.machine().startswith("arm") or platform.platform().startswith('win')
    )


def test_mcstas_reduction_export_to_bytestream(
    reduced_data: NMXReducedDataGroup,
) -> None:
    """Test export method."""
    import h5py
    import numpy as np
    import scipp as sc

    data_fields = [
        'NXdetector',
        'NXsample',
        'NXsource',
        'NXinstrument',
        'definition',
        'name',
    ]

    with io.BytesIO() as bio:
        if not _is_bitshuffle_available():
            # bitshuffle does not build correctly on Windows and ARM machines
            # We are keeping this test here to catch when it builds correctly
            # in the future.
            with pytest.warns(
                DeprecationWarning, match='Please use ``export_as_nxlauetof`` instead.'
            ):
                with pytest.warns(UserWarning, match='bitshuffle.h5'):
                    export_as_nexus(reduced_data, bio)
        else:
            with pytest.warns(
                DeprecationWarning, match='Please use ``export_as_nxlauetof`` instead.'
            ):
                export_as_nexus(reduced_data, bio)

        with h5py.File(bio, 'r') as f:
            assert 'NMX_data' in f
            nmx_data: h5py.Group = f.require_group('NMX_data')
            for field in data_fields:
                assert field in nmx_data

            nx_detector = nmx_data.require_group('NXdetector')
            assert np.all(
                nx_detector['fast_axis'][()] == reduced_data['fast_axis'].values
            )
            assert np.all(
                nx_detector['slow_axis'][()] == reduced_data['slow_axis'].values
            )
            assert np.all(
                nx_detector['origin'][()] == reduced_data['origin_position'].values
            )

            instrument_data = nmx_data.require_group('NXinstrument')
            assert (
                instrument_data['proton_charge'][()]
                == reduced_data['proton_charge'].value
            )

            det1_data = instrument_data.require_group('detector_1')
            assert np.all(det1_data['counts'][()] == reduced_data['counts'].values)
            assert np.all(
                det1_data['pixel_id'][()] == reduced_data['counts'].coords['id'].values
            )
            assert np.all(
                det1_data['t_bin'][()] == reduced_data['counts'].coords['t'].values
            )

            nx_sample = nmx_data.require_group('NXsample')
            sample_name: bytes = nx_sample['name'][()]
            assert sample_name.decode() == reduced_data['sample_name'].value

            nx_source = nmx_data.require_group('NXsource')
            assert (
                nx_source['distance'][()]
                == sc.norm(reduced_data['source_position']).value
            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: RUF100, E402, I

import importlib.metadata

try:
    __version__ = importlib.metadata.version("essnmx")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

from .mcstas import NMXMcStasWorkflow

__all__ = ["NMXMcStasWorkflow"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import pathlib
import subprocess
import time
from contextlib import contextmanager
from enum import Enum

import h5py
import pydantic
import pytest
import scipp as sc
import scippnexus as snx
from scipp.testing import assert_identical

from ess.nmx._executable_helper import (
    InputConfig,
    OutputConfig,
    ReductionConfig,
    WorkflowConfig,
    build_reduction_argument_parser,
    reduction_config_from_args,
)
from ess.nmx.configurations import TimeBinCoordinate, TimeBinUnit, to_command_arguments
from ess.nmx.executables import reduction
from ess.nmx.types import Compression, NMXLauetof


def _build_arg_list_from_pydantic_instance(*instances: pydantic.BaseModel) -> list[str]:
    args = {}
    for instance in instances:
        args.update(instance.model_dump(mode='python'))
    args = {f"--{k.replace('_', '-')}": v for k, v in args.items() if v is not None}

    arg_list = []
    for k, v in args.items():
        if not isinstance(v, bool):
            arg_list.append(k)
            if isinstance(v, list):
                arg_list.extend(str(item) for item in v)
            elif isinstance(v, Enum):
                arg_list.append(v.name)
            else:
                arg_list.append(str(v))
        elif v is True:
            arg_list.append(k)

    return arg_list


def _default_config() -> ReductionConfig:
    """Helper to create a default ReductionConfig instance."""
    return ReductionConfig(
        inputs=InputConfig(input_file=['']),
        workflow=WorkflowConfig(),
        output=OutputConfig(),
    )


def _check_non_default_config(testing_config: ReductionConfig) -> None:
    """Helper to check that all values in the config are non-default."""
    default_config = _default_config()
    testing_children = testing_config._children
    default_children = default_config._children
    for testing_child, default_child in zip(
        testing_children, default_children, strict=True
    ):
        testing_model = testing_child.model_dump(mode='python')
        default_model = default_child.model_dump(mode='python')
        for key, testing_value in testing_model.items():
            if key == 'tof_lookup_table_file_path':
                # This value may be None or default, so we skip the check.
                continue
            default_value = default_model[key]
            assert testing_value != default_value, (
                f"Value for '{key}' is default: {testing_value}"
            )


def test_reduction_config() -> None:
    """Test ReductionConfig argument parsing."""
    # Build config instances with non-default values.
    input_options = InputConfig(
        input_file=['test-input.h5'],
        swmr=True,
        detector_ids=[0, 1, 2, 3],
        iter_chunk=True,
        chunk_size_pulse=10,
        chunk_size_events=100000,
    )
    workflow_options = WorkflowConfig(
        nbins=100,
        min_time_bin=10,
        max_time_bin=100_000,
        time_bin_coordinate=TimeBinCoordinate.event_time_offset,
        time_bin_unit=TimeBinUnit.us,
        tof_simulation_num_neutrons=700_000,
        tof_simulation_max_wavelength=5.0,
        tof_simulation_min_wavelength=1.0,
        tof_simulation_min_ltotal=140.0,
        tof_simulation_max_ltotal=200.0,
        tof_simulation_seed=12345,
    )
    output_options = OutputConfig(
        output_file='test-output.h5',
        compression=Compression.NONE,
        verbose=True,
        skip_file_output=True,
        overwrite=True,
    )
    expected_config = ReductionConfig(
        inputs=input_options, workflow=workflow_options, output=output_options
    )
    # Check if all values are non-default.
    _check_non_default_config(expected_config)

    # Build argument list manually, not using `to_command_arguments` to test it.
    arg_list = _build_arg_list_from_pydantic_instance(
        input_options, workflow_options, output_options
    )
    assert arg_list == to_command_arguments(config=expected_config, one_line=False)

    # Parse arguments and build config from them.
    parser = build_reduction_argument_parser()
    args = parser.parse_args(arg_list)
    config = reduction_config_from_args(args)
    assert expected_config == config


@pytest.fixture(scope="session")
def small_nmx_nexus_path():
    """Fixture to provide the path to the small NMX NeXus file."""
    from ess.nmx.data import get_small_nmx_nexus

    return get_small_nmx_nexus()


def _check_output_file(output_file_path: pathlib.Path, nbins: int):
    detector_names = [f'detector_panel_{i}' for i in range(3)]
    mandatory_fields = (
        'data',
        'distance',
        'fast_axis',
        'slow_axis',
        'origin',
        'x_pixel_size',
        'y_pixel_size',
        'origin',
    )
    with snx.File(output_file_path, 'r') as f:
        # Test
        assert f['entry/instrument/name'][()] == 'NMX'
        for name in detector_names:
            det_gr = f[f'entry/instrument/{name}']
            assert det_gr is not None
            toa_edges = det_gr['time_of_flight'][()]
            assert len(toa_edges) == nbins
            assert all(field_name in det_gr for field_name in mandatory_fields)


def test_executable_runs(small_nmx_nexus_path, tmp_path: pathlib.Path):
    """Test that the executable runs and returns the expected output."""
    output_file = tmp_path / "output.h5"
    assert not output_file.exists()

    nbins = 20  # Small number of bins for testing.
    # The output has 1280x1280 pixels per detector per time bin.
    commands = (
        'essnmx-reduce',
        '--input-file',
        small_nmx_nexus_path,
        '--nbins',
        str(nbins),
        '--output-file',
        output_file.as_posix(),
    )
    # Validate that all commands are strings and contain no unsafe characters
    result = subprocess.run(  # noqa: S603 - We are not accepting arbitrary input here.
        commands, text=True, capture_output=True, check=False
    )
    assert result.returncode == 0
    assert output_file.exists()
    _check_output_file(output_file, nbins=nbins)


@contextmanager
def known_warnings():
    with pytest.warns(RuntimeWarning, match="No crystal rotation*"):
        yield


@pytest.fixture
def temp_output_file(tmp_path: pathlib.Path):
    output_file_path = tmp_path / "scipp_output.h5"
    yield output_file_path
    if output_file_path.exists():
        output_file_path.unlink()


@pytest.fixture
def reduction_config(
    small_nmx_nexus_path: pathlib.Path, temp_output_file: pathlib.Path
) -> ReductionConfig:
    input_config = InputConfig(input_file=[small_nmx_nexus_path.as_posix()])
    # Compression option is not default (NONE) but
    # the actual default compression option, BITSHUFFLE_LZ4,
    # only properly works in linux so we set it to NONE here
    # for convenience of testing on all platforms.
    output_config = OutputConfig(
        output_file=temp_output_file.as_posix(),
        compression=Compression.NONE,
        skip_file_output=True,  # No need to write output file for most tests.
    )
    return ReductionConfig(inputs=input_config, output=output_config)


def _retrieve_one_hist(results: NMXLauetof) -> sc.DataArray:
    """Helper to retrieve the first DataArray from the results dictionary."""
    da = results.instrument.detectors['detector_panel_0'].data
    if not isinstance(da, sc.DataArray):
        raise TypeError("Histogram is not a DataArray.")
    return da


def test_reduction_default_settings(reduction_config: ReductionConfig) -> None:
    # Only check if reduction runs without errors with default settings.
    with known_warnings():
        reduction(config=reduction_config)


def test_reduction_only_number_of_time_bins(reduction_config: ReductionConfig) -> None:
    reduction_config.workflow.nbins = 20
    with known_warnings():
        hist = _retrieve_one_hist(reduction(config=reduction_config))

    # Check that the number of time bins is as expected.
    assert len(hist.coords['tof']) == 21  # nbins + 1 edges


def test_histogram_event_time_offset(reduction_config: ReductionConfig) -> None:
    reduction_config.workflow.nbins = 20
    reduction_config.workflow.time_bin_coordinate = TimeBinCoordinate.event_time_offset
    with known_warnings():
        hist = _retrieve_one_hist(reduction(config=reduction_config))

    # Check that the number of time bins is as expected.
    assert len(hist.coords['event_time_offset']) == 21  # nbins + 1 edges
    # Check if the histogram result is reasonable
    zero = sc.scalar(0.0, unit='counts', dtype='float32', variance=0.0)
    assert bool(hist.data.sum() > zero)


def test_histogram_invalid_min_max_raises(reduction_config: ReductionConfig) -> None:
    reduction_config.workflow.min_time_bin = 120
    reduction_config.workflow.max_time_bin = 100
    with pytest.raises(ValueError, match='Cannot build a time bin edges coordinate'):
        with known_warnings():
            reduction(config=reduction_config)


def test_histogram_invalid_min_max_raises_eto(
    reduction_config: ReductionConfig,
) -> None:
    reduction_config.workflow.time_bin_coordinate = TimeBinCoordinate.event_time_offset
    reduction_config.workflow.min_time_bin = 50
    reduction_config.workflow.max_time_bin = 40
    with pytest.raises(ValueError, match='Cannot build a time bin edges coordinate'):
        with known_warnings():
            reduction(config=reduction_config)


@pytest.mark.parametrize(
    argnames="t_coord",
    argvalues=[TimeBinCoordinate.time_of_flight, TimeBinCoordinate.event_time_offset],
)
def test_histogram_out_of_range_min_warns(
    reduction_config: ReductionConfig, t_coord: TimeBinCoordinate
) -> None:
    reduction_config.workflow.time_bin_coordinate = t_coord
    reduction_config.workflow.nbins = 20
    reduction_config.workflow.min_time_bin = 1_000
    reduction_config.workflow.max_time_bin = 2_000
    with pytest.warns(UserWarning, match='is bigger than all'):
        with known_warnings():
            results = reduction(config=reduction_config)

    for histogram in results.instrument.detectors.values():
        assert isinstance(histogram.data, sc.DataArray)
        da = histogram.data
        assert_identical(
            da.data.sum(), sc.scalar(0.0, unit='counts', dtype='float32', variance=0.0)
        )


@pytest.mark.parametrize(
    argnames="t_coord",
    argvalues=[TimeBinCoordinate.time_of_flight, TimeBinCoordinate.event_time_offset],
)
def test_histogram_out_of_range_max_warns(
    reduction_config: ReductionConfig, t_coord: TimeBinCoordinate
) -> None:
    reduction_config.workflow.time_bin_coordinate = t_coord
    reduction_config.workflow.nbins = 10
    reduction_config.workflow.min_time_bin = -1
    reduction_config.workflow.max_time_bin = 0
    with pytest.warns(UserWarning, match='is smaller than all'):
        with known_warnings():
            results = reduction(config=reduction_config)

    for det in results.instrument.detectors.values():
        da = det.data
        assert isinstance(da, sc.DataArray)
        assert_identical(
            da.data.sum(), sc.scalar(0.0, unit='counts', dtype='float32', variance=0.0)
        )


@pytest.fixture
def tof_lut_file_path(tmp_path: pathlib.Path):
    """Fixture to provide the path to the small NMX NeXus file."""
    from dataclasses import is_dataclass

    from ess.reduce.time_of_flight import TimeOfFlightLookupTable

    from ess.nmx.workflows import initialize_nmx_workflow

    # Simply use the default workflow for testing.
    workflow = initialize_nmx_workflow(config=WorkflowConfig())
    tof_lut: TimeOfFlightLookupTable = workflow.compute(TimeOfFlightLookupTable)

    # Change the tof range a bit for testing.
    if isinstance(tof_lut, sc.DataArray):
        tof_lut *= 2
    elif is_dataclass(tof_lut):
        tof_lut.array *= 2
    else:
        raise TypeError("Unexpected type for TOF lookup table.")

    lut_file_path = tmp_path / "nmx_tof_lookup_table.h5"
    tof_lut.save_hdf5(lut_file_path.as_posix())
    yield lut_file_path
    if lut_file_path.exists():
        lut_file_path.unlink()


def test_reduction_with_tof_lut_file(
    reduction_config: ReductionConfig, tof_lut_file_path: pathlib.Path
) -> None:
    # Make sure the config uses no TOF lookup table file initially.
    assert reduction_config.workflow.tof_lookup_table_file_path is None
    with known_warnings():
        default_results = reduction(config=reduction_config)

    # Update config to use the TOF lookup table file.
    reduction_config.workflow.tof_lookup_table_file_path = tof_lut_file_path.as_posix()
    with known_warnings():
        results = reduction(config=reduction_config)

    default_hists = [det.data for det in default_results.instrument.detectors.values()]
    hists = [det.data for det in results.instrument.detectors.values()]

    for default_hist, hist in zip(default_hists, hists, strict=True):
        assert isinstance(default_hist, sc.DataArray)
        assert isinstance(hist, sc.DataArray)
        tof_edges_default = default_hist.coords['tof']
        tof_edges = hist.coords['tof']
        assert_identical(default_hist.data, hist.data)
        assert_identical(tof_edges_default * 2, tof_edges)


def test_reduction_succeed_when_skipping_evenif_output_file_exists(
    reduction_config: ReductionConfig, temp_output_file: pathlib.Path
) -> None:
    # Make sure the file exists
    temp_output_file.touch(exist_ok=True)
    # Make sure the file output is skipped.
    reduction_config.output.skip_file_output = True

    # Adjust workflow setting to finish fast.
    reduction_config.workflow.nbins = 2
    reduction_config.workflow.time_bin_coordinate = TimeBinCoordinate.event_time_offset
    with known_warnings():
        reduction(config=reduction_config)


def test_reduction_fails_fast_if_output_file_exists(
    reduction_config: ReductionConfig, temp_output_file: pathlib.Path
) -> None:
    # Make sure the file exists
    temp_output_file.touch()
    # Make sure file output is NOT skipped.
    reduction_config.output.skip_file_output = False

    start = time.time()
    with pytest.raises(FileExistsError):
        reduction(config=reduction_config)
    finish = time.time()

    # Check if the `reduction` call fails within 1 second.
    # There is no special reason why it is 1 second.
    # It should just fail as fast as possible.
    assert finish - start < 1


def test_reduction_compression_gzip(
    reduction_config: ReductionConfig, tmp_path: pathlib.Path
) -> None:
    reduction_config.output.skip_file_output = False
    reduction_config.workflow.nbins = 5  # For faster test
    file_paths: dict[Compression, pathlib.Path] = {}

    for compress_mode in (Compression.NONE, Compression.GZIP):
        reduction_config.output.compression = compress_mode
        cur_file_path = tmp_path / f'compress_{compress_mode}_output.hdf'
        file_paths[compress_mode] = cur_file_path
        assert not cur_file_path.exists()
        reduction_config.output.output_file = cur_file_path.as_posix()
        # Running the whole reduction instead of only saving the file on purpose.
        with known_warnings():
            reduction(config=reduction_config)
        assert cur_file_path.exists()

    assert (
        file_paths[Compression.NONE].stat().st_size
        > file_paths[Compression.GZIP].stat().st_size
    )
    with h5py.File(file_paths[Compression.NONE]) as file:
        for i in range(3):
            assert file[f'entry/instrument/detector_panel_{i}/data'].chunks is None

    with h5py.File(file_paths[Compression.GZIP]) as file:
        for i in range(3):
            data_path = f'entry/instrument/detector_panel_{i}/data'
            assert file[data_path].chunks == (1280, 1280, 1)
            assert file[data_path].compression == 'gzip'
            assert file[data_path].compression_opts == 4


try:
    # Just checking availability
    import bitshuffle.h5  # noqa: F401
except ImportError:
    BITSHUFFLE_AVAILABLE = False
else:
    BITSHUFFLE_AVAILABLE = True


@pytest.mark.skipif(
    not BITSHUFFLE_AVAILABLE,
    reason="Bitshuffle is not available in this environment.",
)
def test_reduction_compression_bitshuffle_smaller_than_gzip(
    reduction_config: ReductionConfig, tmp_path: pathlib.Path
) -> None:
    reduction_config.output.skip_file_output = False
    reduction_config.workflow.nbins = 5  # For faster test
    file_paths: dict[Compression, pathlib.Path] = {}
    total_times: dict[Compression, pathlib.Path] = {}

    for compress_mode in (Compression.GZIP, Compression.BITSHUFFLE_LZ4):
        reduction_config.output.compression = compress_mode
        cur_file_path = tmp_path / f'compress_{compress_mode}_output.hdf'
        file_paths[compress_mode] = cur_file_path
        assert not cur_file_path.exists()
        reduction_config.output.output_file = cur_file_path.as_posix()
        # Running the whole reduction instead of only saving the file on purpose.
        with known_warnings():
            start = time.time()
            reduction(config=reduction_config)
            end = time.time()

        assert cur_file_path.exists()
        total_times[compress_mode] = end - start

    # GZIP is expected to have better compression ratio than BITSHUFFLE
    assert (
        file_paths[Compression.BITSHUFFLE_LZ4].stat().st_size
        > file_paths[Compression.GZIP].stat().st_size
    )
    # BITSHUFFLE is expected to be faster than GZIP
    assert total_times[Compression.BITSHUFFLE_LZ4] < total_times[Compression.GZIP]

    with h5py.File(file_paths[Compression.GZIP]) as file:
        for i in range(3):
            data_path = f'entry/instrument/detector_panel_{i}/data'
            assert file[data_path].chunks == (1280, 1280, 1)
            assert file[data_path].compression == 'gzip'

    with h5py.File(file_paths[Compression.BITSHUFFLE_LZ4]) as file:
        for i in range(3):
            data_path = f'entry/instrument/detector_panel_{i}/data'
            assert file[data_path].chunks == (1280, 1280, 1)
            # For some reason it doesn't write the compression.
            # so we check the filter instead.
            # assert file[data_path].compression == 'bitshuffle'
            assert '32008' in file[data_path]._filters


@pytest.mark.skipif(
    BITSHUFFLE_AVAILABLE,
    reason="Bitshuffle is available in this environment so it won't fall back.",
)
def test_reduction_compression_bitshuffle_fall_back_to_gzip(
    reduction_config: ReductionConfig, temp_output_file: pathlib.Path
) -> None:
    reduction_config.output.skip_file_output = False
    reduction_config.workflow.nbins = 5  # For faster test
    reduction_config.output.compression = Compression.BITSHUFFLE_LZ4
    reduction_config.output.output_file = temp_output_file.as_posix()

    with known_warnings():
        with pytest.warns(UserWarning, match='bitshuffle.h5'):
            reduction(config=reduction_config)

    with h5py.File(temp_output_file) as file:
        for i in range(3):
            data_path = f'entry/instrument/detector_panel_{i}/data'
            assert file[data_path].chunks == (1280, 1280, 1)
            assert file[data_path].compression == 'gzip'


def test_reduction_duplicated_path_raises(reduction_config: ReductionConfig) -> None:
    # Run with two files with same names.
    reduction_config.inputs.input_file = reduction_config.inputs.input_file * 2
    with pytest.raises(
        ValueError, match=r'Duplicated file paths or pattern found.*small_nmx_nexus.hdf'
    ):
        reduction(config=reduction_config)



"""
DIALS .refl file loader

This loads msgpack-type DIALS reflection files, without having DIALS or
cctbx in the python environment.

Note: All modern .refl files are at time of writing msgpack-based. Some
much older files might be in pickle format, which this doesn't read.

Adapted from Nick Cavendish of the DIALS team.
"""

import functools
import logging
import operator
import os
import struct
from collections.abc import Iterable
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import IO, cast

import msgpack
import numpy as np


@dataclass
class Shoebox:
    panel: int
    bbox: tuple[int]
    data: np.ndarray | None = None
    mask: np.ndarray | None = None
    background: np.ndarray | None = None


def _decode_raw_numpy(dtype, shape: int | Iterable = 1):
    """
    Decoding a column that maps straight to a numpy array.

    Args:
        dtype: The numpy dtype for the array
        shape:
            The shape of a single item. Either an int, or a collection
            of ints, in C-array order (row major)
    """
    # Convert to a shape tuple
    if isinstance(shape, int):
        shape = (shape,)
    else:
        shape = tuple(shape)

    def _decode_specific(data, copy):
        num_items, raw = data
        array = np.frombuffer(raw, dtype=dtype)

        if shape != (1,):
            item_width = functools.reduce(operator.mul, shape)
            if len(raw) % item_width != 0:
                raise AssertionError(
                    "Raw data length %s not divisible by item width %s",
                    len(raw),
                    item_width,
                )
            if num_items * item_width != len(array):
                raise AssertionError(
                    "(Num items) %s * (item width) %s != (raw data length) %s",
                    num_items,
                    item_width,
                    len(raw),
                )
            array = array.reshape(num_items, *shape)
        if copy:
            return np.copy(array)
        return array

    return _decode_specific


def _decode_shoeboxes(data: list, copy) -> Iterable[Shoebox]:
    # Shoebox is float
    num_items, raw = data
    shoeboxes: list[Shoebox | None] = []
    pos = 0
    while pos < len(raw):
        sbox_header_fmt = "<IiiiiiiB"
        sb_info = struct.unpack_from(sbox_header_fmt, raw, pos)
        pos += struct.calcsize(sbox_header_fmt)
        panel = sb_info[0]
        bbox = sb_info[1:7]
        data_present = sb_info[7]
        shoebox = {"panel": panel, "bbox": bbox}
        if data_present:
            bbox_size = (bbox[5] - bbox[4], bbox[3] - bbox[2], bbox[1] - bbox[0])
            data_size = (bbox_size[0] * bbox_size[1] * bbox_size[2]) * 4
            # Read three sets of data: data, mask and background
            shoebox["data"] = np.frombuffer(
                raw[pos : pos + data_size], dtype=np.float32
            ).reshape(bbox_size)
            pos += data_size
            shoebox["mask"] = np.frombuffer(
                raw[pos : pos + data_size], dtype=np.int32
            ).reshape(bbox_size)
            pos += data_size
            shoebox["background"] = np.frombuffer(
                raw[pos : pos + data_size], dtype=np.float32
            ).reshape(bbox_size)
            pos += data_size
            if copy:
                shoebox["data"] = np.copy(shoebox["data"])
                shoebox["mask"] = np.copy(shoebox["mask"])
                shoebox["background"] = np.copy(shoebox["background"])

        # Although this is technically a divergence,
        # return None instead of an empty shoebox
        if not data_present and all(x == 0 for x in bbox) and panel == 0:
            shoeboxes.append(None)
        else:
            shoeboxes.append(Shoebox(**shoebox))
    if len(shoeboxes) != num_items:
        raise AssertionError(
            f"Mismatch of shoebox lengths: {len(shoeboxes)} != {num_items}"
        )
    return np.array(shoeboxes, dtype=np.object_)


_reftable_decoders = {
    "bool": _decode_raw_numpy(bool),
    "int": _decode_raw_numpy(np.int32),
    "double": _decode_raw_numpy(np.double),
    "int6": _decode_raw_numpy(np.int32, shape=6),
    "std::size_t": _decode_raw_numpy(np.uint64),
    "vec3<double>": _decode_raw_numpy(np.double, shape=3),
    "cctbx::miller::index<>": _decode_raw_numpy(np.int32, shape=3),
    "Shoebox<>": _decode_shoeboxes,
    "vec2<double>": _decode_raw_numpy(np.double, shape=2),
    "mat3<double>": _decode_raw_numpy(np.double, shape=(3, 3)),
    # "std::string": _decode_wip, # - string writing broken; dials/dials#1858
}


def decode_column(column_entry, copy):
    """Decode a single column value"""
    datatype, data = column_entry

    converter = _reftable_decoders.get(datatype)
    if not converter:
        logging.warning(
            "Data type '%s' does not have a converter; cannot read", datatype
        )
        return None
    return converter(data, copy=copy)


def _get_unpacked(stream_or_path: str | IO | bytes | os.PathLike):
    """Works out the logic to pass a stream/pathlike to msgpack"""
    try:
        logging.INFO(type(stream_or_path))
        path = os.fspath(cast(str, stream_or_path))
        is_fspathlike = True
    except (TypeError, ValueError):
        path = stream_or_path
        is_fspathlike = isinstance(stream_or_path, str)

    if is_fspathlike:
        with open(path, "rb") as f:
            un = msgpack.Unpacker(f, strict_map_key=False)
            return un.unpack()
    else:
        un = msgpack.Unpacker(stream_or_path, strict_map_key=False)
        return un.unpack()


def loads(data: bytes, copy=False):
    """
    Load a DIALS msgpack-encoded .refl file.

    Args:
        data: bytes data, already read from the file.
        copy: Should the data be copied into writable numpy arrays.

    Returns: See .load(stream_or_path)
    """
    return load(BytesIO(data), copy)


def load(stream_or_path: IO | Path | os.PathLike, copy=False) -> dict:
    """
    Load a DIALS msgpack-encoded .refl file

    Args:
        stream_or_path: The filename or data to load
        copy:
            Should the data be copied. This will cause more memory usage
            whilst loading the raw data.

    Returns:

        A dictionary with each column in the reflection table. If there
        is an identifier mapping as part of the reflection table, then
        this is returned as an extra 'experiment_identifier' column.
        All columns except Shoeboxes are returned as numpy arrays,
        except Shoebox columns, which are returned as Dataclass objects
        which contain the portions of data from the file.

        With copy=False, all numpy arrays are pointing against the raw
        memory returned by msgpack, which means they are read-only.
        With copy=True, an immediate copy is done. This causes memory
        usage to double while loading, but the created numpy arrays own
        their own memory.
    """
    root_data = _get_unpacked(stream_or_path)

    if not root_data[0] == "dials::af::reflection_table":
        raise ValueError("Does not appear to be a dials reflection table file")
    if not root_data[1] == 1:
        raise ValueError(
            f"reflection_table data is version {root_data[1]}."
            "Only Version 1 is understood"
        )
    refdata = root_data[2]

    rows = refdata["nrows"]
    identifiers = refdata["identifiers"]
    data = refdata["data"]

    decoded_data = {
        name: decode_column(value, copy=copy) for name, value in data.items()
    }

    # Filter out empty (unknown) columns
    decoded_data = {k: v for k, v in decoded_data.items() if v is not None}

    # Cross-check the columns are the expected lengths
    for name, column in decoded_data.items():
        if len(column) != rows:
            logging.warning(
                "Warning: Mismatch of column lengths: "
                "[%s] is [%d] instead of expected [%s]",
                name,
                len(column),
                rows,
            )

    # Make an "identifiers" column
    if "id" in decoded_data and identifiers:
        decoded_data["experiment_identifier"] = [
            identifiers[x] for x in decoded_data["id"] if x > 0
        ]

    return decoded_data



# flake8: noqa: E501

no_detectors = """
SPLIT 999 COMPONENT Xtal = Single_crystal(
                order = 1,
                p_transmit=0.001,
        reflections = "Rubredoxin.lau",
        xwidth = XtalSize_width,
                yheight = XtalSize_height,
                zdepth = XtalSize_depth,
                mosaic = XtalMosaicity,
                delta_d_d=1e-4)
   AT (0, 0, deltaz) RELATIVE PREVIOUS
   ROTATED (XtalPhiX,XtalPhiY, XtalPhiZ) RELATIVE armSample
 EXTEND %{
        if (!SCATTERED) {ABSORB;}
  %}

COMPONENT Sphere = PSD_monitor_4PI(
    nx = 360, ny = 360, filename = "4pi", radius = 0.2,
    restore_neutron = 1)

AT (0, 0, deltaz) RELATIVE armSample
"""

two_detectors_two_filenames = """
COMPONENT nD_Mantid_0 = Monitor_nD(
        options ="mantid square x limits=[0 0.512] bins=1280 y limits=[0 0.512] bins=1280, neutron pixel min=1 t, list all neutrons",
    xmin = 0,
    xmax = 0.512,
    ymin = 0,
    ymax = 0.512,
    restore_neutron = 1,
    filename = "bank01_events.dat")
        AT (-0.25, -0.25, 0.29) RELATIVE armSample
        ROTATED (0, 0, 0) RELATIVE armSample

COMPONENT nD_Mantid_1 = Monitor_nD(
        options ="mantid square x limits=[0 0.512] bins=1280 y limits=[0 0.512] bins=1280, neutron pixel min=2000000 t, list all neutrons",
    xmin = 0,
    xmax = 0.512,
    ymin = 0,
    ymax = 0.512,
    restore_neutron = 1,
    filename = "bank02_events.dat")
  AT (-0.29, -0.25, 0.25) RELATIVE armSample
  ROTATED (0, 90, 0) RELATIVE armSample
"""

one_detector_no_filename = """
COMPONENT nD_Mantid_2 = Monitor_nD(
        options ="mantid square x limits=[0 0.512] bins=1280 y limits=[0 0.512] bins=1280, neutron pixel min=2000000 t, list all neutrons",
    xmin = 0,
    xmax = 0.512,
    ymin = 0,
    ymax = 0.512,
    restore_neutron = 1,
  AT (-0.29, -0.25, 0.25) RELATIVE armSample
  ROTATED (0, 90, 0) RELATIVE armSample
"""

two_detectors_same_filename = """
COMPONENT nD_Mantid_0 = Monitor_nD(
        options ="mantid square x limits=[0 0.512] bins=1280 y limits=[0 0.512] bins=1280, neutron pixel min=1 t, list all neutrons",
    xmin = 0,
    xmax = 0.512,
    ymin = 0,
    ymax = 0.512,
    restore_neutron = 1,
    filename = "bank01_events.dat")
        AT (-0.25, -0.25, 0.29) RELATIVE armSample
        ROTATED (0, 0, 0) RELATIVE armSample

COMPONENT nD_Mantid_1 = Monitor_nD(
        options ="mantid square x limits=[0 0.512] bins=1280 y limits=[0 0.512] bins=1280, neutron pixel min=2000000 t, list all neutrons",
    xmin = 0,
    xmax = 0.512,
    ymin = 0,
    ymax = 0.512,
    restore_neutron = 1,
    filename = "bank01_events.dat")
  AT (-0.29, -0.25, 0.25) RELATIVE armSample
  ROTATED (0, 90, 0) RELATIVE armSample
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from .types import MaximumCounts

default_parameters = {MaximumCounts: 10000}


def NMXMcStasWorkflow():
    import sciline as sl

    from .load import providers as loader_providers
    from .reduction import (
        calculate_maximum_toa,
        calculate_minimum_toa,
        format_nmx_reduced_data,
        proton_charge_from_event_counts,
        raw_event_probability_to_counts,
        reduce_raw_event_probability,
    )
    from .xml import read_mcstas_geometry_xml

    return sl.Pipeline(
        (
            *loader_providers,
            calculate_maximum_toa,
            calculate_minimum_toa,
            read_mcstas_geometry_xml,
            proton_charge_from_event_counts,
            reduce_raw_event_probability,
            raw_event_probability_to_counts,
            format_nmx_reduced_data,
        ),
        params=default_parameters,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import argparse
import logging
import pathlib
from collections.abc import Callable
from functools import partial

import sciline as sl
import scipp as sc
from ess.reduce.streaming import (
    EternalAccumulator,
    MaxAccumulator,
    MinAccumulator,
    StreamProcessor,
)

from ..types import Compression
from . import NMXMcStasWorkflow
from .load import (
    mcstas_weight_to_probability_scalefactor,
    raw_event_data_chunk_generator,
)
from .nexus import (
    _export_detector_metadata_as_nxlauetof,
    _export_reduced_data_as_nxlauetof,
    _export_static_metadata_as_nxlauetof,
)
from .streaming import calculate_number_of_chunks
from .types import (
    DetectorIndex,
    DetectorName,
    FilePath,
    MaximumCounts,
    MaximumProbability,
    MaximumTimeOfArrival,
    McStasWeight2CountScaleFactor,
    MinimumTimeOfArrival,
    NMXDetectorMetadata,
    NMXExperimentMetadata,
    NMXRawDataMetadata,
    NMXReducedCounts,
    NMXReducedDataGroup,
    PixelIds,
    RawEventProbability,
    TimeBinSteps,
)
from .xml import McStasInstrument


def _build_metadata_streaming_processor_helper() -> Callable[
    [sl.Pipeline], StreamProcessor
]:
    return partial(
        StreamProcessor,
        dynamic_keys=(RawEventProbability,),
        target_keys=(NMXRawDataMetadata,),
        accumulators={
            MaximumProbability: MaxAccumulator,
            MaximumTimeOfArrival: MaxAccumulator,
            MinimumTimeOfArrival: MinAccumulator,
        },
    )


def _build_final_streaming_processor_helper() -> Callable[
    [sl.Pipeline], StreamProcessor
]:
    return partial(
        StreamProcessor,
        dynamic_keys=(RawEventProbability,),
        target_keys=(NMXReducedDataGroup,),
        accumulators={NMXReducedCounts: EternalAccumulator},
    )


def calculate_raw_data_metadata(
    *detector_ids: DetectorIndex | DetectorName,
    wf: sl.Pipeline,
    chunk_size: int = 10_000_000,
    logger: logging.Logger | None = None,
) -> NMXRawDataMetadata:
    # Stream processor building helper
    scalefactor_stream_processor = _build_metadata_streaming_processor_helper()
    metadata_wf = wf.copy()
    # Loop over the detectors
    file_path = metadata_wf.compute(FilePath)
    raw_data_metadatas = {}

    for detector_i in detector_ids:
        temp_wf = metadata_wf.copy()
        if isinstance(detector_i, str):
            temp_wf[DetectorName] = detector_i
        else:
            temp_wf[DetectorIndex] = detector_i

        detector_name = temp_wf.compute(DetectorName)
        max_chunk_id = calculate_number_of_chunks(
            temp_wf.compute(FilePath),
            detector_name=detector_name,
            chunk_size=chunk_size,
        )
        # Build the stream processor
        processor = scalefactor_stream_processor(temp_wf)
        for i_da, da in enumerate(
            raw_event_data_chunk_generator(
                file_path=file_path, detector_name=detector_name, chunk_size=chunk_size
            )
        ):
            if any(da.sizes.values()) == 0:
                continue
            else:
                results = processor.add_chunk({RawEventProbability: da})
            if logger is not None:
                logger.info(
                    "[{%s}/{%s}] Processed chunk for {%s}",
                    i_da + 1,
                    max_chunk_id,
                    detector_name,
                )

        raw_data_metadatas[detector_i] = results[NMXRawDataMetadata]

    # We take the min/maximum values of the scale factor
    # We are doing it manually because it is not possible to update parameters
    # in the workflow that stream processor uses.
    min_toa = min(dg.min_toa for dg in raw_data_metadatas.values())
    max_toa = max(dg.max_toa for dg in raw_data_metadatas.values())
    max_probability = max(dg.max_probability for dg in raw_data_metadatas.values())

    return NMXRawDataMetadata(
        min_toa=min_toa, max_toa=max_toa, max_probability=max_probability
    )


def reduction(
    *,
    input_file: pathlib.Path,
    output_file: pathlib.Path,
    chunk_size: int = 10_000_000,
    nbins: int = 50,
    max_counts: int | None = None,
    detector_ids: list[int | str],
    compression: Compression = Compression.BITSHUFFLE_LZ4,
    wf: sl.Pipeline | None = None,
    logger: logging.Logger | None = None,
    toa_min_max_prob: tuple[float] | None = None,
) -> None:
    wf = wf.copy() if wf is not None else NMXMcStasWorkflow()
    wf[FilePath] = input_file
    # Set static info
    wf[McStasInstrument] = wf.compute(McStasInstrument)

    if not toa_min_max_prob:
        # Calculate parameters for data reduction
        data_metadata = calculate_raw_data_metadata(
            *detector_ids, wf=wf, logger=logger, chunk_size=chunk_size
        )
        if logger is not None:
            logger.info("Metadata retrieved: %s", data_metadata)

        toa_bin_edges = sc.linspace(
            dim='t',
            start=data_metadata.min_toa,
            stop=data_metadata.max_toa,
            num=nbins + 1,
        )
        scale_factor = mcstas_weight_to_probability_scalefactor(
            max_counts=wf.compute(MaximumCounts),
            max_probability=data_metadata.max_probability,
        )
    else:
        if logger is not None:
            logger.info("Metadata given: %s", toa_min_max_prob)
        toa_min = sc.scalar(toa_min_max_prob[0], unit='s')
        toa_max = sc.scalar(toa_min_max_prob[1], unit='s')
        prob_max = sc.scalar(toa_min_max_prob[2])
        toa_bin_edges = sc.linspace(dim='t', start=toa_min, stop=toa_max, num=nbins + 1)
        scale_factor = mcstas_weight_to_probability_scalefactor(
            max_counts=wf.compute(MaximumCounts),
            max_probability=prob_max,
        )

    if max_counts:
        scale_factor = mcstas_weight_to_probability_scalefactor(
            max_counts=MaximumCounts(max_counts),
            max_probability=data_metadata.max_probability,
        )
    else:
        scale_factor = mcstas_weight_to_probability_scalefactor(
            max_counts=wf.compute(MaximumCounts),
            max_probability=data_metadata.max_probability,
        )
    # Compute metadata and make the skeleton output file
    experiment_metadata = wf.compute(NMXExperimentMetadata)
    detector_metas = []
    for detector_i in range(3):
        temp_wf = wf.copy()
        temp_wf[DetectorIndex] = detector_i
        detector_metas.append(temp_wf.compute(NMXDetectorMetadata))

    if logger is not None:
        logger.info("Exporting metadata into the output file %s", output_file)

    _export_static_metadata_as_nxlauetof(
        experiment_metadata=experiment_metadata,
        output_file=output_file,
        # Arbitrary metadata falls into ``entry`` group as a variable.
        mcstas_weight2count_scale_factor=scale_factor,
    )
    _export_detector_metadata_as_nxlauetof(*detector_metas, output_file=output_file)
    # Compute histogram
    final_wf = wf.copy()
    # Set the scale factor and time bin edges
    final_wf[McStasWeight2CountScaleFactor] = scale_factor
    final_wf[TimeBinSteps] = toa_bin_edges

    file_path = final_wf.compute(FilePath)
    final_stream_processor = _build_final_streaming_processor_helper()
    # Loop over the detectors
    result_list = []
    for detector_i in detector_ids:
        temp_wf = final_wf.copy()
        if isinstance(detector_i, str):
            temp_wf[DetectorName] = detector_i
        else:
            temp_wf[DetectorIndex] = detector_i
        # Set static information as parameters
        detector_name = temp_wf.compute(DetectorName)
        temp_wf[PixelIds] = temp_wf.compute(PixelIds)
        max_chunk_id = calculate_number_of_chunks(
            file_path, detector_name=detector_name, chunk_size=chunk_size
        )

        # Build the stream processor
        processor = final_stream_processor(temp_wf)
        for i_da, da in enumerate(
            raw_event_data_chunk_generator(
                file_path=file_path, detector_name=detector_name, chunk_size=chunk_size
            )
        ):
            if any(da.sizes.values()) == 0:
                continue
            else:
                results = processor.add_chunk({RawEventProbability: da})
            if logger is not None:
                logger.info(
                    "[{%s}/{%s}] Processed chunk for {%s}",
                    i_da + 1,
                    max_chunk_id,
                    detector_name,
                )

        result = results[NMXReducedDataGroup]
        result_list.append(result)
        if logger is not None:
            logger.info("Appending reduced data into the output file %s", output_file)

        _export_reduced_data_as_nxlauetof(
            result,
            output_file=output_file,
            compress_counts=(compression == Compression.NONE),
        )
    from ess.nmx.reduction import merge_panels

    return merge_panels(*result_list)


def _add_mcstas_args(parser: argparse.ArgumentParser) -> None:
    mcstas_arg_group = parser.add_argument_group("McStas Data Reduction Options")
    mcstas_arg_group.add_argument(
        "--max_counts",
        type=int,
        default=None,
        help="Maximum Counts",
    )
    mcstas_arg_group.add_argument(
        "--chunk_size",
        type=int,
        default=10_000_000,
        help="Chunk size for processing (number of events per chunk)",
    )


def build_reduction_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Command line arguments for the NMX reduction. "
        "It assumes 14 Hz pulse speed."
    )
    input_arg_group = parser.add_argument_group("Input Options")
    input_arg_group.add_argument(
        "--input_file", type=str, help="Path to the input file", required=True
    )
    input_arg_group.add_argument(
        "--nbins",
        type=int,
        default=50,
        help="Number of TOF bins",
    )
    input_arg_group.add_argument(
        "--detector_ids",
        type=int,
        nargs="+",
        default=[0, 1, 2],
        help="Detector indices to process",
    )

    output_arg_group = parser.add_argument_group("Output Options")
    output_arg_group.add_argument(
        "--output_file",
        type=str,
        default="scipp_output.h5",
        help="Path to the output file",
    )
    output_arg_group.add_argument(
        "--compression",
        type=str,
        default=Compression.BITSHUFFLE_LZ4.name,
        choices=[compression_key.name for compression_key in Compression],
        help="Compress option of reduced output file. Default: BITSHUFFLE_LZ4",
    )
    output_arg_group.add_argument(
        "--verbose", "-v", action="store_true", help="Increase output verbosity"
    )

    return parser


def main() -> None:
    from .._executable_helper import build_logger

    parser = build_reduction_arg_parser()
    _add_mcstas_args(parser)
    args = parser.parse_args()

    input_file = pathlib.Path(args.input_file).resolve()
    output_file = pathlib.Path(args.output_file).resolve()

    logger = build_logger(args)

    wf = NMXMcStasWorkflow()
    reduction(
        input_file=input_file,
        output_file=output_file,
        chunk_size=args.chunk_size,
        nbins=args.nbins,
        max_counts=args.max_counts,
        detector_ids=args.detector_ids,
        compression=Compression[args.compression],
        logger=logger,
        wf=wf,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import re
from collections.abc import Generator

import scipp as sc
import scippnexus as snx

from .types import (
    CrystalRotation,
    DetectorBankPrefix,
    DetectorIndex,
    DetectorName,
    FilePath,
    MaximumCounts,
    MaximumProbability,
    MaximumTimeOfArrival,
    McStasWeight2CountScaleFactor,
    MinimumTimeOfArrival,
    NMXDetectorMetadata,
    NMXExperimentMetadata,
    NMXRawDataMetadata,
    NMXRawEventCountsDataGroup,
    PixelIds,
    RawEventProbability,
)
from .xml import McStasInstrument, read_mcstas_geometry_xml


def detector_name_from_index(index: DetectorIndex) -> DetectorName:
    return f'nD_Mantid_{getattr(index, "value", index)}'


def load_event_data_bank_name(
    detector_name: DetectorName, file_path: FilePath
) -> DetectorBankPrefix:
    '''Finds the filename associated with a detector'''
    with snx.File(file_path) as file:
        description = file['entry1/instrument/description'][()]
        for bank_name, det_names in bank_names_to_detector_names(description).items():
            if detector_name in det_names:
                return DetectorBankPrefix(bank_name.partition('.')[0])
    raise KeyError(
        f"{DetectorBankPrefix.__name__} cannot be found for "
        f"{DetectorName.__name__} from the file {FilePath.__name__}"
    )


def _exclude_zero_events(data: sc.Variable) -> sc.Variable:
    """Exclude events with zero counts from the data.

    McStas can add extra event lines containing 0,0,0,0,0,0
    These lines should not be included so we skip it.
    """
    data = data[(data != sc.scalar(0.0, unit=data.unit)).any(dim="dim_1")]
    return data


def _wrap_raw_event_data(data: sc.Variable) -> RawEventProbability:
    data = data.rename_dims({'dim_0': 'event'})
    data = _exclude_zero_events(data)
    try:
        event_da = sc.DataArray(
            coords={
                'id': sc.array(
                    dims=['event'],
                    values=data['dim_1', 4].values,
                    dtype='int64',
                    unit=None,
                ),
                't': sc.array(dims=['event'], values=data['dim_1', 5].values, unit='s'),
            },
            data=sc.array(
                dims=['event'], values=data['dim_1', 0].values, unit='counts'
            ),
        )
    except IndexError:
        event_da = sc.DataArray(
            coords={
                'id': sc.array(
                    dims=['event'],
                    values=data['dim_1', 1].values,
                    dtype='int64',
                    unit=None,
                ),
                't': sc.array(dims=['event'], values=data['dim_1', 2].values, unit='s'),
            },
            data=sc.array(
                dims=['event'], values=data['dim_1', 0].values, unit='counts'
            ),
        )

    return RawEventProbability(event_da)


def load_raw_event_data(
    file_path: FilePath, *, detector_name: DetectorName, bank_prefix: DetectorBankPrefix
) -> RawEventProbability:
    """Retrieve events from the nexus file.

    Parameters
    ----------
    file_path:
        Path to the nexus file
    detector_name:
        Name of the detector to load
    bank_prefix:
        Prefix identifying the event data array containing the events of the detector
        If None, the bank name is determined automatically from the detector name.

    """
    if bank_prefix is None:
        bank_prefix = load_event_data_bank_name(detector_name, file_path)
    bank_name = f'{bank_prefix}_dat_list_p_x_y_n_id_t'
    with snx.File(file_path, 'r') as f:
        root = f["entry1/data"]
        (bank_name,) = (name for name in root.keys() if bank_name in name)
        data = root[bank_name]["events"][()]
        return _wrap_raw_event_data(data)


def _check_chunk_size(chunk_size: int) -> None:
    if 0 < chunk_size < 10_000_000:
        import warnings

        warnings.warn(
            "The chunk size may be too small < 10_000_000.\n"
            "Consider increasing the chunk size for better performance.\n"
            "Hint: NMX typically expect ~10^8 bins as reduced data.",
            UserWarning,
            stacklevel=2,
        )


def _check_maximum_chunk_size(d_slices: tuple[slice, ...]) -> None:
    """Check the maximum size of the slices."""
    max_chunk_size = max(
        (d_slice.stop - d_slice.start) / d_slice.step for d_slice in d_slices
    )
    _check_chunk_size(max_chunk_size)


def _validate_chunk_size(chunk_size: int) -> None:
    """Validate the chunk size."""
    if not isinstance(chunk_size, int):
        raise TypeError("Chunk size must be an integer.")
    if chunk_size < -1:
        raise ValueError("Invalid chunk size. It should be -1(for all) or > 0.")


def raw_event_data_chunk_generator(
    file_path: FilePath,
    *,
    detector_name: DetectorName,
    bank_prefix: DetectorBankPrefix | None = None,
    chunk_size: int = 0,  # Number of rows to read at a time
) -> Generator[RawEventProbability, None, None]:
    """Chunk events from the nexus file.

    Parameters
    ----------
    file_path:
        Path to the nexus file
    detector_name:
        Name of the detector to load
    pixel_ids:
        Pixel ids to generate the data array with the events
    chunk_size:
        Number of rows to read at a time.
        If 0, chunk slice is determined automatically by the ``iter_chunks``.
        Note that it only works if the dataset is already chunked.

    Yields
    ------
    RawEventProbability:
        Data array containing the events of the detector.

    Raises
    ------
    ValueError:
        If the chunk size is not valid. (>= -1)
    TypeError:
        If the chunk size is not an integer.
    Warning
        If the chunk size is too small (< 10_000_000).

    """
    _check_chunk_size(chunk_size)
    _validate_chunk_size(chunk_size)

    # Find the data bank name associated with the detector
    bank_prefix = load_event_data_bank_name(
        detector_name=detector_name, file_path=file_path
    )
    bank_name = f'{bank_prefix}_dat_list_p_x_y_n_id_t'
    with snx.File(file_path, 'r') as f:
        root = f["entry1/data"]
        (bank_name,) = (name for name in root.keys() if bank_name in name)

    with snx.File(file_path, 'r') as f:
        root = f["entry1/data"]
        dset = root[bank_name]["events"]
        if chunk_size == 0:
            # dset.dataset.iter_chunks() yields (dim_0_slice, dim_1_slice)
            dim_0_slices = tuple(dim0_sl for dim0_sl, _ in dset.dataset.iter_chunks())
            # Only checking maximum chunk size
            # since the last chunk may be smaller than the rest of the chunks
            _check_maximum_chunk_size(dim_0_slices)
            for dim_0_slice in dim_0_slices:
                da = _wrap_raw_event_data(dset["dim_0", dim_0_slice])
                yield da
        elif chunk_size == -1:
            yield _wrap_raw_event_data(dset[()])
        else:
            num_events = dset.shape[0]
            for start in range(0, num_events, chunk_size):
                data = dset["dim_0", start : start + chunk_size]
                yield _wrap_raw_event_data(data)


def load_crystal_rotation(
    file_path: FilePath, instrument: McStasInstrument
) -> CrystalRotation:
    """Retrieve crystal rotation from the file.

    Raises
    ------
    KeyError
        If the crystal rotation is not found in the file.

    """
    with snx.File(file_path, 'r') as file:
        param_keys = tuple(f"entry1/simulation/Param/XtalPhi{key}" for key in "XYZ")
        if not all(key in file for key in param_keys):
            raise KeyError(
                f"Crystal rotations [{', '.join(param_keys)}] not found in file."
            )
        return CrystalRotation(
            sc.vector(
                value=[file[param_key][...] for param_key in param_keys],
                unit=instrument.simulation_settings.angle_unit,
            )
        )


def maximum_probability(da: RawEventProbability) -> MaximumProbability:
    """Find the maximum probability in the data."""
    return MaximumProbability(da.data.max())


def mcstas_weight_to_probability_scalefactor(
    max_counts: MaximumCounts, max_probability: MaximumProbability
) -> McStasWeight2CountScaleFactor:
    """Calculate the scale factor to convert McStas weights to counts.

    max_counts * (probabilities / max_probability)

    Parameters
    ----------
    max_counts:
        The maximum number of counts after scaling the event counts.

    scale_factor:
        The scale factor to convert McStas weights to counts

    """

    return McStasWeight2CountScaleFactor(
        sc.scalar(max_counts, unit="counts") / max_probability
    )


def bank_names_to_detector_names(description: str) -> dict[str, list[str]]:
    """Associates event data names with the names of the detectors
    where the events were detected"""

    detector_component_regex = (
        # Start of the detector component definition, contains the detector name.
        # r'^COMPONENT (?P<detector_name>.*) = Monitor_nD\(\n'
        r'^COMPONENT (?P<detector_name>.*) = (Monitor_nD|Union_abs_logger_nD)\(\n'
        # Some uninteresting lines, we're looking for 'filename'.
        # Make sure no new component begins.
        r'(?:(?!COMPONENT)(?!filename)(?:.|\s))*'
        # The line that defines the filename of the file that stores the
        # events associated with the detector.
        r'(?:filename = \"(?P<bank_name>[^\"]*)\")?'
    )
    matches = re.finditer(detector_component_regex, description, re.MULTILINE)
    bank_names_to_detector_names = {}
    for m in matches:
        bank_names_to_detector_names.setdefault(
            # If filename was not set for the detector the filename for the
            # event data defaults to the name of the detector.
            m.group('bank_name') or m.group('detector_name'),
            [],
        ).append(m.group('detector_name'))
    return bank_names_to_detector_names


def load_experiment_metadata(
    instrument: McStasInstrument, crystal_rotation: CrystalRotation
) -> NMXExperimentMetadata:
    """Load the experiment metadata from the McStas file."""
    return NMXExperimentMetadata(
        sc.DataGroup(
            crystal_rotation=crystal_rotation, **instrument.experiment_metadata()
        )
    )


def load_detector_metadata(
    instrument: McStasInstrument, detector_name: DetectorName
) -> NMXDetectorMetadata:
    """Load the detector metadata from the McStas file."""
    return NMXDetectorMetadata(
        sc.DataGroup(**instrument.detector_metadata(detector_name))
    )


def load_mcstas(
    *,
    da: RawEventProbability,
    experiment_metadata: NMXExperimentMetadata,
    detector_metadata: NMXDetectorMetadata,
) -> NMXRawEventCountsDataGroup:
    return NMXRawEventCountsDataGroup(
        sc.DataGroup(weights=da, **experiment_metadata, **detector_metadata)
    )


def retrieve_pixel_ids(
    instrument: McStasInstrument, detector_name: DetectorName
) -> PixelIds:
    """Retrieve the pixel IDs for a given detector."""
    return PixelIds(instrument.pixel_ids(detector_name))


def retrieve_raw_data_metadata(
    min_toa: MinimumTimeOfArrival,
    max_toa: MaximumTimeOfArrival,
    max_probability: MaximumProbability,
) -> NMXRawDataMetadata:
    """Retrieve the metadata of the raw data."""
    return NMXRawDataMetadata(
        min_toa=min_toa, max_toa=max_toa, max_probability=max_probability
    )


providers = (
    retrieve_raw_data_metadata,
    read_mcstas_geometry_xml,
    detector_name_from_index,
    load_event_data_bank_name,
    load_raw_event_data,
    maximum_probability,
    mcstas_weight_to_probability_scalefactor,
    retrieve_pixel_ids,
    load_crystal_rotation,
    load_mcstas,
    load_experiment_metadata,
    load_detector_metadata,
)



import enum
from dataclasses import dataclass, field
from typing import Literal, NewType

import h5py
import numpy as np
import scipp as sc
import scippnexus as snx
from ess.reduce.time_of_flight.types import TofLookupTable
from scippneutron.metadata import RadiationProbe, SourceType

from ._display_helper import to_datagroup


class Compression(enum.StrEnum):
    """Compression type of the output file.

    These options are written as enum for future extensibility.
    """

    NONE = 'NONE'
    GZIP = 'GZIP'
    BITSHUFFLE_LZ4 = 'BITSHUFFLE_LZ4'


TofSimulationMinWavelength = NewType("TofSimulationMinWavelength", sc.Variable)
"""Minimum wavelength for tof simulation to calculate look up table."""

TofSimulationMaxWavelength = NewType("TofSimulationMaxWavelength", sc.Variable)
"""Maximum wavelength for tof simulation to calculate look up table."""


class ControlMode(enum.StrEnum):
    """Control mode of counting.

    Based on the NXlauetof definition of ``control`` (NXmonitor) field.
    """

    monitor = 'monitor'
    """Count to a preset value based on received monitor counts."""
    timer = 'timer'
    """Count to a preset value based on clock time"""


def _unit_matrix() -> sc.Variable:
    return sc.spatial.linear_transform(
        value=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        unit="dimensionless",
    )


def _uniform_unit_cell_length() -> sc.Variable:
    return sc.vector([1.0, 1.0, 1.0], unit='dimensionless')


def _cube_unit_cell_angle() -> sc.Variable:
    return sc.vector([90.0, 90.0, 90.0], unit='deg')


@dataclass(kw_only=True)
class NMXSampleMetadata:
    nx_class = snx.NXsample

    crystal_rotation: sc.Variable
    name: str
    position: sc.Variable
    # Temporarily hardcoding some values
    # TODO: Remove hardcoded values
    orientation_matrix: sc.Variable = field(default_factory=_unit_matrix)
    unit_cell_length: sc.Variable = field(default_factory=_uniform_unit_cell_length)
    unit_cell_angle: sc.Variable = field(default_factory=_cube_unit_cell_angle)

    @property
    def unit_cell(self) -> sc.Variable:
        """a, b, c, alpha, beta, gamma."""

        return np.concat([self.unit_cell_length.values, self.unit_cell_angle.values])

    def __write_to_nexus_group__(self, group: h5py.Group):
        cr_field = snx.create_field(group, 'crystal_rotation', self.crystal_rotation)
        cr_field.attrs['long_name'] = 'crystal rotation in Phi (XYZ)'
        snx.create_field(group, 'name', self.name)
        snx.create_field(group, 'position', self.position)
        snx.create_field(group, 'orientation_matrix', self.orientation_matrix)
        unit_cell = snx.create_field(group, 'unit_cell', self.unit_cell)
        unit_cell.attrs['length-unit'] = str(self.unit_cell_length.unit)
        unit_cell.attrs['angle-unit'] = str(self.unit_cell_angle.unit)


@dataclass(kw_only=True)
class NMXSourceMetadata:
    nx_class = snx.NXsource

    position: sc.Variable
    """Position of the source (from the sample)."""

    # These three fields are matching fields as ``scippneutron.metadata.Source``.
    # However, NMX needs to store `position` as a vector,
    # not only the name, type and probe
    # essnmx cannot use ``scippneutron.metadata.Source`` as it is.
    # We will need to implement unpacking function for vector scalar value.
    # Therefore we decided not to use the ``scippneutron.metadata.Source`` for now
    # but the ``NMXSourceMetadata`` 's ``source_type`` and ``probe`` fields
    # have the same Enum types as ``scippneutron.metadata.Source``.
    name: Literal['European Spallation Source'] = "European Spallation Source"
    source_type: SourceType = SourceType.SpallationNeutronSource
    probe: RadiationProbe = RadiationProbe.Neutron

    @property
    def distance(self) -> sc.Variable:
        return sc.norm(self.position)

    def __write_to_nexus_group__(self, group: h5py.Group):
        snx.create_field(group, 'name', self.name)
        snx.create_field(group, 'type', self.source_type.value)
        distance = snx.create_field(group, 'distance', self.distance)
        distance.attrs['position'] = self.position.values
        snx.create_field(group, 'probe', self.probe.value)


def _zero_float_count() -> sc.Variable:
    return sc.scalar(0.0, unit='count')


@dataclass(kw_only=True)
class NMXMonitorMetadata:
    nx_class = snx.NXmonitor
    data: sc.DataArray
    """Monitor counts."""

    @property
    def time_of_flight(self) -> sc.Variable:
        return self.data.coords[self.tof_bin_coord]

    tof_bin_coord: str = field(
        default='tof',
        metadata={
            "description": "Name of the time-of-flight coordinate "
            "in the monitor histogram."
        },
    )
    mode: ControlMode = field(
        default=ControlMode.monitor,
        metadata={"description": "Mode of counting. One of `monitor` or `timer`."},
    )
    preset: sc.Variable = field(
        default_factory=_zero_float_count,
        metadata={"description": "Preset value of counting for the `mode`."},
    )

    def __write_to_nexus_group__(self, group: h5py.Group):
        group.attrs['axes'] = self.data.dims
        group.attrs['tof_bin_coord'] = self.tof_bin_coord
        snx.create_field(group, 'mode', str(self.mode))
        snx.create_field(group, 'preset', self.preset)
        data_field = snx.create_field(group, 'data', self.data.data)
        data_field.attrs['signal'] = 1
        data_field.attrs['primary'] = 1
        snx.create_field(group, 'time_of_flight', self.time_of_flight)


@dataclass(kw_only=True)
class NMXDetectorMetadata:
    nx_class = snx.NXdetector

    detector_name: str
    x_pixel_size: sc.Variable
    y_pixel_size: sc.Variable
    origin: sc.Variable
    """Center of the detector panel."""
    fast_axis: sc.Variable
    """Inner most dimension if the data is sorted by detector number.

    The index of the fast axis changes fast along the detector number.

    i.e. When detector numbers grows: ``0, 1, 2, 3, 4, 5, 6, ...``
    and the size of the fast axis is ``3``,
    the fast axis index will be: ``0, 1, 2, 0, 1, 2, 0 ...``
    for each detector number.

    """
    fast_axis_dim: str
    slow_axis: sc.Variable
    """Outer most dimension if the data is sorted by detector number.

    The index of the slow axis changes slowly along the detector number.

    i.e. When detector numbers grows: ``0, 1, 2, 3, 4, 5, 6, ...``
    and the size of the fast axis is ``3``,
    the slow axis index will be: ``0, 0, 0, 1, 1, 1, 2, ...``
    for each detector number.

    """
    slow_axis_dim: str
    distance: sc.Variable
    first_pixel_position: sc.Variable
    """First pixel position with respect to the sample.

    Additional field for DIALS. It should be a 3D vector.
    """
    # TODO: Remove hardcoded values
    polar_angle: sc.Variable = field(default_factory=lambda: sc.scalar(0, unit='deg'))
    azimuthal_angle: sc.Variable = field(
        default_factory=lambda: sc.scalar(0, unit='deg')
    )

    def __write_to_nexus_group__(self, group: h5py.Group):
        snx.create_field(group, 'x_pixel_size', self.x_pixel_size)
        snx.create_field(group, 'y_pixel_size', self.y_pixel_size)
        origin = snx.create_field(group, 'origin', self.origin)
        origin.attrs['first_pixel_position'] = self.first_pixel_position.values
        fast_axis = snx.create_field(group, 'fast_axis', self.fast_axis)
        fast_axis.attrs['dim'] = self.fast_axis_dim
        slow_axis = snx.create_field(group, 'slow_axis', self.slow_axis)
        slow_axis.attrs['dim'] = self.slow_axis_dim
        snx.create_field(group, 'distance', self.distance)
        snx.create_field(group, 'polar_angle', self.polar_angle)
        snx.create_field(group, 'azimuthal_angle', self.azimuthal_angle)


@dataclass(kw_only=True)
class NMXReducedDetector:
    """Reduced Detector data and metadata container.

    In an output file, all metadata fields are stored on the same level as the `data`.
    However, in this reduced detector data container, the `data` and `metadata` are
    separated with an extra hierarchy.
    It is because the `data` needs more control how to be stored,
    i.e. compression option.
    Also, the histogram may need chunk-wise processing
    and therefore metadata may need to be written in advance so that
    the `data` can be appended to the existing `NXdetector` HDF5 Group.

    """

    data: sc.DataArray | None = None
    """3D Histogram of the detector counts or its place holder."""
    metadata: NMXDetectorMetadata
    """NMX Detector metadata."""


@dataclass(kw_only=True)
class NMXInstrument:
    nx_class = snx.NXinstrument

    detectors: sc.DataGroup[NMXReducedDetector]
    name: str = "NMX"
    source: NMXSourceMetadata


@dataclass(kw_only=True)
class NMXProgram:
    nx_class = 'NXprogram'

    program: str = 'essnmx'

    def __write_to_nexus_group__(self, group: h5py.Group):
        from ess.nmx import __version__ as essnmxversion

        prog = snx.create_field(group, 'program', self.program)
        prog.attrs['version'] = essnmxversion


@dataclass(kw_only=True)
class NMXLauetof:
    nx_class = "NXlauetof"

    control: NMXMonitorMetadata
    definitions: Literal['NXlauetof'] = 'NXlauetof'
    instrument: NMXInstrument
    sample: NMXSampleMetadata
    lookup_table: TofLookupTable | None = None
    reducer: NMXProgram = field(default_factory=NMXProgram)
    "Information of the reduction software."

    def to_datagroup(self) -> sc.DataGroup:
        return to_datagroup(self)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import scipp as sc

from .types import (
    MaximumTimeOfArrival,
    McStasWeight2CountScaleFactor,
    MinimumTimeOfArrival,
    NMXDetectorMetadata,
    NMXExperimentMetadata,
    NMXReducedCounts,
    NMXReducedDataGroup,
    NMXReducedProbability,
    PixelIds,
    ProtonCharge,
    RawEventProbability,
    TimeBinSteps,
)


def calculate_minimum_toa(da: RawEventProbability) -> MinimumTimeOfArrival:
    """Calculate the minimum time of arrival from the data."""
    return MinimumTimeOfArrival(da.coords['t'].min())


def calculate_maximum_toa(da: RawEventProbability) -> MaximumTimeOfArrival:
    """Calculate the maximum time of arrival from the data."""
    return MaximumTimeOfArrival(da.coords['t'].max())


def proton_charge_from_event_counts(da: NMXReducedCounts) -> ProtonCharge:
    """Make up the proton charge from the event counts.

    Proton charge is proportional to the number of neutrons,
    which is proportional to the number of events.
    The scale factor is manually chosen based on previous results
    to be convenient for data manipulation in the next steps.
    It is derived this way since
    the protons are not part of McStas simulation,
    and the number of neutrons is not included in the result.

    Parameters
    ----------
    event_da:
        The event data

    """
    # Arbitrary number to scale the proton charge
    return ProtonCharge(sc.scalar(1 / 10_000, unit='dimensionless') * da.data.sum())


def reduce_raw_event_probability(
    da: RawEventProbability, pixel_ids: PixelIds, time_bin_step: TimeBinSteps
) -> NMXReducedProbability:
    return NMXReducedProbability(da.group(pixel_ids).hist(t=time_bin_step))


def raw_event_probability_to_counts(
    da: NMXReducedProbability,
    scale_factor: McStasWeight2CountScaleFactor,
) -> NMXReducedCounts:
    return NMXReducedCounts(da * scale_factor)


def format_nmx_reduced_data(
    da: NMXReducedCounts,
    proton_charge: ProtonCharge,
    experiment_metadata: NMXExperimentMetadata,
    detector_metadata: NMXDetectorMetadata,
) -> NMXReducedDataGroup:
    """Bin time of arrival data into ``time_bin_step`` bins."""

    return NMXReducedDataGroup(
        sc.DataGroup(
            counts=da,
            proton_charge=proton_charge,
            **experiment_metadata,
            **detector_metadata,
        )
    )


def _concat_or_same(
    obj: list[sc.Variable | sc.DataArray], dim: str
) -> sc.Variable | sc.DataArray:
    first = obj[0]
    # instrument.to_coords in bin_time_of_arrival adds a panel coord to some fields,
    # even if it has only length 1. If this is the case we concat, even if identical.
    # Maybe McStasInstrument.to_coords should be changed to only handle a single
    # panel, and not perform concatenation?
    if all(dim not in o.dims and sc.identical(first, o) for o in obj):
        return first
    return sc.concat(obj, dim)


def merge_panels(*panel: NMXReducedDataGroup) -> NMXReducedDataGroup:
    """Merge a list of panels by concatenating along the 'panel' dimension."""
    keys = panel[0].keys()
    if not all(p.keys() == keys for p in panel):
        raise ValueError("All panels must have the same keys.")
    return NMXReducedDataGroup(
        sc.DataGroup(
            {key: _concat_or_same([p[key] for p in panel], 'panel') for key in keys}
        )
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Scipp contributors (https://github.com/scipp)
import pathlib
from contextlib import contextmanager

import pytest
from scipp.testing.assertions import assert_allclose, assert_identical

from ess.nmx._nxlauetof_io import load_essnmx_nxlauetof
from ess.nmx.configurations import InputConfig, OutputConfig, ReductionConfig
from ess.nmx.executables import reduction
from ess.nmx.types import Compression


@pytest.fixture
def temp_output_file(tmp_path: pathlib.Path):
    output_file_path = tmp_path / "scipp_output.h5"
    yield output_file_path
    if output_file_path.exists():
        output_file_path.unlink()


@pytest.fixture
def reduction_config(temp_output_file: pathlib.Path) -> ReductionConfig:
    from ess.nmx.data import get_small_nmx_nexus

    input_config = InputConfig(input_file=[get_small_nmx_nexus().as_posix()])
    output_config = OutputConfig(
        output_file=temp_output_file.as_posix(),
        compression=Compression.NONE,
        skip_file_output=False,
    )
    return ReductionConfig(inputs=input_config, output=output_config)


@contextmanager
def known_warnings():
    with pytest.warns(RuntimeWarning, match="No crystal rotation*"):
        yield


def test_loaded_data_same_as_in_memory_result(
    reduction_config: ReductionConfig,
) -> None:
    with known_warnings():
        result = reduction(config=reduction_config)
    original_result_dg = result.to_datagroup()

    # Adjust original result to be same as expected loaded data group.
    original_result_dg.pop('lookup_table')
    original_positions = {}
    detectors = original_result_dg['instrument']['detectors']
    for det_name, det in detectors.items():
        # Removing coordinates that are not kept in the file or reconstructed.
        det['data'].coords.pop('Ltotal')
        det['data'].coords.pop('detector_number')
        det['data'].coords.pop('x_pixel_offset')
        det['data'].coords.pop('y_pixel_offset')
        # Saving position coordinate to compare them by allclose instead of eq.
        original_positions[det_name] = det['data'].coords.pop('position')

    loaded_dg = load_essnmx_nxlauetof(reduction_config.output.output_file)

    loaded_detector_positions = {}
    for det_name, loaded_det in loaded_dg['instrument']['detectors'].items():
        loaded_detector_positions[det_name] = loaded_det['data'].coords.pop('position')

    assert_identical(loaded_dg, original_result_dg)
    # Using the x_pixel_size of the first panel to get absolute tolerance.
    pixel_size = next(iter(detectors.values()))['metadata']['x_pixel_size']
    atol = pixel_size / 10.0
    for det_name, original_position in original_positions.items():
        loaded_position = loaded_detector_positions[det_name]
        assert_allclose(original_position, loaded_position, atol=atol)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pathlib
import sys
from collections.abc import Generator

import pytest
import scipp as sc
import scippnexus as snx
from scipp.testing import assert_allclose, assert_identical

from ess.nmx import NMXMcStasWorkflow
from ess.nmx.data import get_small_mcstas
from ess.nmx.mcstas.load import bank_names_to_detector_names, load_crystal_rotation
from ess.nmx.mcstas.types import (
    DetectorBankPrefix,
    DetectorIndex,
    FilePath,
    NMXRawEventCountsDataGroup,
)

sys.path.insert(0, str(pathlib.Path(__file__).resolve().parent))
from mcstas_description_examples import (
    no_detectors,
    one_detector_no_filename,
    two_detectors_same_filename,
    two_detectors_two_filenames,
)


def check_nmxdata_properties(
    dg: NMXRawEventCountsDataGroup, fast_axis, slow_axis
) -> None:
    assert isinstance(dg, sc.DataGroup)
    assert_allclose(dg['fast_axis'], fast_axis, atol=sc.scalar(0.005))
    assert_identical(dg['slow_axis'], slow_axis)


def check_scalar_properties_mcstas_3(dg: NMXRawEventCountsDataGroup):
    """Test helper for NMXData loaded from McStas 3.

    Expected numbers are hard-coded based on the sample file.
    """
    assert_identical(dg['crystal_rotation'], sc.vector([0, 0, 0], unit='deg'))
    assert_identical(dg['sample_position'], sc.vector(value=[0, 0, 0], unit='m'))
    assert_identical(
        dg['source_position'], sc.vector(value=[-0.53123, 0.0, -157.405], unit='m')
    )
    assert dg['sample_name'] == sc.scalar("sampleMantid")


@pytest.mark.parametrize(
    ('detector_index', 'fast_axis', 'slow_axis'),
    [
        # Expected values are provided by the IDS
        # based on the simulation settings of the sample file.
        (0, (1.0, 0.0, -0.01), (0.0, 1.0, 0.0)),
        (1, (-0.01, 0.0, -1.0), (0.0, 1.0, 0.0)),
        (2, (0.01, 0.0, 1.0), (0.0, 1.0, 0.0)),
    ],
)
def test_file_reader_mcstas3(detector_index, fast_axis, slow_axis) -> None:
    file_path = get_small_mcstas()

    pl = NMXMcStasWorkflow()
    pl[FilePath] = file_path
    pl[DetectorIndex] = detector_index
    dg, bank = pl.compute((NMXRawEventCountsDataGroup, DetectorBankPrefix)).values()

    entry_path = f"entry1/data/{bank}_dat_list_p_x_y_n_id_t"
    with snx.File(file_path) as file:
        raw_data = file[entry_path]["events"][()]
        data_length = raw_data.sizes['dim_0']

    check_scalar_properties_mcstas_3(dg)
    assert dg['weights'].sizes['event'] == data_length
    check_nmxdata_properties(dg, sc.vector(fast_axis), sc.vector(slow_axis))


@pytest.fixture(params=[get_small_mcstas])
def tmp_mcstas_file(
    tmp_path: pathlib.Path,
    request: pytest.FixtureRequest,
) -> Generator[pathlib.Path, None, None]:
    import os
    import shutil

    original_file_path = request.param()

    tmp_file = tmp_path / pathlib.Path('file.h5')
    shutil.copy(original_file_path, tmp_file)
    yield tmp_file
    os.remove(tmp_file)


def test_file_reader_mcstas_additional_fields(tmp_mcstas_file: pathlib.Path) -> None:
    """Check if additional fields names do not break the loader."""
    import h5py

    entry_path = "entry1/data/bank01_events_dat_list_p_x_y_n_id_t"
    new_entry_path = entry_path + '_L'

    with h5py.File(tmp_mcstas_file, 'r+') as file:
        dataset = file[entry_path]
        del file[entry_path]
        file[new_entry_path] = dataset

    pl = NMXMcStasWorkflow()
    pl[FilePath] = str(tmp_mcstas_file)
    pl[DetectorIndex] = 0
    dg = pl.compute(NMXRawEventCountsDataGroup)

    assert isinstance(dg, sc.DataGroup)


@pytest.fixture
def rotation_mission_tmp_file(tmp_mcstas_file: pathlib.Path) -> pathlib.Path:
    import h5py

    param_keys = tuple(f"entry1/simulation/Param/XtalPhi{key}" for key in "XYZ")

    # Remove the rotation parameters from the file.
    with h5py.File(tmp_mcstas_file, 'a') as file:
        for key in param_keys:
            del file[key]

    return tmp_mcstas_file


def test_missing_rotation(rotation_mission_tmp_file: FilePath) -> None:
    with pytest.raises(KeyError, match="XtalPhiX"):
        load_crystal_rotation(rotation_mission_tmp_file, None)
        # McStasInstrument is not used due to error in the file.


def test_bank_names_to_detector_names_two_detectors():
    res = bank_names_to_detector_names(two_detectors_two_filenames)
    assert len(res) == 2
    assert all(len(v) == 1 for v in res.values())


def test_bank_names_to_detector_names_same_filename():
    res = bank_names_to_detector_names(two_detectors_same_filename)
    assert len(res) == 1
    assert all(len(v) == 2 for v in res.values())


def test_bank_names_to_detector_names_no_detectors():
    res = bank_names_to_detector_names(no_detectors)
    assert len(res) == 0


def test_bank_names_to_detector_names_no_filename():
    res = bank_names_to_detector_names(one_detector_no_filename)
    assert len(res) == 1
    ((bank, (detector,)),) = res.items()
    assert bank == detector
